(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

(* Adapted from Ocaml/toplevel/toploop.ml for MLF by Boris Yakobowski *)

open Constraints
open ConstraintsGeneration
open SyntaxErrors


let first_line = ref true
let got_eof = ref false;;

let read_input_default prompt buffer len =
  output_string stdout prompt; flush stdout;
  let i = ref 0 in
  try
    while true do
      if !i >= len then raise Exit;
      let c = input_char stdin in
      buffer.[!i] <- c;
      incr i;
      if c = '\n' then raise Exit;
    done;
    (!i, false)
  with
  | End_of_file ->
      (!i, true)
  | Exit ->
      (!i, false)

let read_interactive_input = ref read_input_default

let refill_lexbuf buffer len =
  if !got_eof then (got_eof := false; 0) else begin
    let prompt =
(*      if !Clflags.noprompt then "" *)
      if !first_line then "# "
      else if !Lexer.in_comment then "* "
      else "  "
    in
    first_line := false;
    let (len, eof) = !read_interactive_input prompt buffer len in
    if eof then begin
      Highlight.echo_eof ();
      if len > 0 then got_eof := true;
      len
    end else
      len
  end

let _ =
  Sys.interactive := true;;

let rec skip_phrase lexbuf =
  try
    match Lexer.token lexbuf with
      Parser.SEMISEMI | Parser.EOF -> ()
    | _ -> skip_phrase lexbuf
  with
    | LexingError (UnterminatedComment _) -> ()
    | LexingError (InvalidChar _) -> skip_phrase lexbuf
;;

(* Maybe introduced a bug. TODO *)
let maybe_skip_phrase lexbuf =
  if Lexing.lexeme lexbuf = ";;" || lexbuf.Lexing.lex_eof_reached
  then ()
  else skip_phrase lexbuf

let wrap parsing_fun lexbuf =
  try
    let ast = parsing_fun Lexer.token lexbuf in
    Parsing.clear_parser();
    ast
  with
  | LexingError (UnterminatedComment _) as exn -> raise exn
  | LexingError (InvalidChar _) as exn ->
      if !Highlight.input_name = "" then skip_phrase lexbuf;
      raise exn
  | (ParsingError _ | Parser.Error) as exn ->
      if !Highlight.input_name = "" then maybe_skip_phrase lexbuf;
      raise exn
;;

let parse_toplevel_phrase = wrap Parser.toplevel_phrase


exception PPerror

open Highlight
open Lexing

let loop ppf =
  Format.fprintf ppf "        MLF version 0.9@.@.";
  let lb = Lexing.from_function refill_lexbuf in
  Highlight.input_name := "";
  Highlight.input_lexbuf := Some lb;
  Sys.catch_break true;
  let env = ref TypingEnv.initial_env in
  while true do
    try
      Lexing.flush_input lb;
      Highlight.reset();
      first_line := true;
      let phr = try parse_toplevel_phrase lb with Exit -> raise PPerror in
      env := TypePhrases.type_phrase ~ppf !env phr
    with
    | End_of_file -> exit 0
    | Sys.Break -> Format.fprintf ppf "Interrupted.@."
    | PPerror -> ()
    | Parser.Error ->
      let loc = Locations.curr_loc lb in
      if !input_name = "";
      then maybe_skip_phrase lb;
      print_loc ppf loc;
      Format.fprintf ppf "Syntax error.@."
    | Failure "TypePhrases" -> ()
    | x -> ErrorsHandling.report_error ppf x
  done


let usage =
"Toplevel loop for typing MLF declarations.
Usage: toplevel [options]"

let () =
  Arg.parse Options.common (fun _ -> ()) usage;

  loop Format.std_formatter
