(*pp camlp4o ./hashtbl_ext.cmo *)

open GraphicTypes
open GraphicTypes.Nodes
open Utils
open Constraints
open OptionsInteractive


let () = #{HashNodes}

let contains_different_nodes l =
  List.length (Utils.unique Nodes.equal l) > 1

type edge = Struct of int option * node * node | Exists of node * node

(* Feeble attempt at printing the nodes in a pleasant order *)
let succs_print c n =
  let exists = Constraints.existentials_on c n in
  try (* Arrow *)
    let n_ar = List.find (fun n' -> !!n'.constr = Constr `Arrow &&
        Nodes.equal (snd (child2 n')) (child1 n)) exists
    in
    let n_c_ar =
      try Some (let e = List.find (fun {edge_nodes = (_, n)} -> n = n_ar)
                  c.constraints#get in fst (e.edge_nodes))
      with Not_found -> None
    and n_c_res =
      let res = fst (child2 n_ar) in
      try Some (let e = List.find (fun {edge_nodes = (_, n)} -> n = res)
                  c.constraints#get in fst (e.edge_nodes))
      with Not_found -> None
    in
    un_option_list [n_c_ar; n_c_res; Some n_ar] @ exists

  with _ ->
    exists @ children n

let dfs_print c = GraphicTypesAux.aux_dfs (succs_print c)


(* Printing of a constraint in Dot format *)
let print_constraint_aux ch ~output ~show_mono ~skip_root c =
  let b1 = Buffer.create 100 and b2 = Buffer.create 100
  and name b n = Printf.bprintf b "n%d" (nid n)
  and is_mono = if !gc_nodes then is_mono_constraint c else (fun _ -> false) in
  let print_node n =
    (* Declaration of the nodes *)
    Printf.bprintf b1 "  %a [%s=\"%s\", style=filled, fillcolor=%s];\n"
      name n
      (if output = Options.LaTeX then "texlbl" else "label")
      (if !OptionsInteractive.print_nodes_id then
         string_of_int (nid n)
       else
         print_gconstr ~output () !!n.constr)
      (match is_mono n, !!n.bound with
         | true, _ -> "white"
         | false, Root -> "green"
         | false, Bound { permissions = p } -> print_permissions p)
    ;

    (* Structure edges *)
    let struct_end = if !show_struct_end then "" else ", dir=none" in
    if contains_different_nodes (children n) &&
       (!sort_edges = false || !show_children_number) then
      iteri (fun i n' -> Printf.bprintf b1
               "  %a -> %a [label=\"%d\", color=black%s];\n"
               name n name n' i struct_end)
        1 (children n)
    else
      List.iter (fun n' -> Printf.bprintf b1
                   "  %a -> %a [color=black%s];\n" name n name n' struct_end)
        (children n);

    (* Binding edges *)
    match !!n.bound with
      | Root -> ()
      | Bound {binder = n'; label = l} ->
          let is_mono = is_mono n in
          if is_mono = false || show_mono || List.mem n c.exists#get then
            Printf.bprintf
              (if c.exists#mem n then b1 else b2)
              "  %a -> %a [color=%s, dir=back];\n"
              name n' name n
              (match is_mono, l with
                 | true, _ -> "grey"
                 | false, Flex -> "blue"
                 | false, Rigid -> "red"
            );
  in
  let r = if skip_root then (child1 c.root) else c.root in
  let b = !!r.bound in
  !!r.bound <- Root;
  (dfs_print c).GraphicTypesAux.generic_dfs ~post:print_node r;
  !!r.bound <- b;
  Printf.fprintf ch "subgraph cluster_graph {\n%s}\n\n%s"
    (Buffer.contents b1) (Buffer.contents b2)
;;

let print_constraints_edges ch c =
  let name ch n = Printf.fprintf ch "n%d" (nid n) in
  let print_edge { edge_type = ct; edge_nodes = (n, n') } = match ct with
    | Constraints.Unif ->
        Printf.fprintf ch
         "  %a -> %a [color=green, dir=both, constraint=false, style=dashed];\n"
          name n name n'
    | Constraints.Inst ->
        Printf.fprintf ch
          "  %a -> %a [color=red, dir=back, style=dashed%s];\n"
          name n' name n
          (if !inst_constrains_height then "" else " ,constraint=false")
  in
  List.iter print_edge c.Constraints.constraints#get;
;;

let print_header ch =
  Printf.fprintf ch "digraph G {\n";
  Printf.fprintf ch "  nodesep=0.35;\n";
  Printf.fprintf ch "  ranksep=0;\n";
  Printf.fprintf ch "  node [fontname=\"../interactive/CODE2000.TTF\", fixedsize=\"true\", width=.45, height=.4];\n";
  if !sort_edges then
    Printf.fprintf ch "  ordering=out;\n";
;;

let print_footer ch =
  Printf.fprintf ch "}\n\n";
;;

let print_constraint ch ?(skip_root=false) ?(show_mono=(!Options.show_mono_edges)) ?(output=(!Options.output)) c =
  print_header ch;
  print_constraint_aux ch ~show_mono ~skip_root ~output c;
  print_constraints_edges ch c;
  print_footer ch;
;;

let to_svg f =
  let temp1, ch1 = Filename.open_temp_file "mlf" ".dot" in
  f ch1;
  close_out ch1;
  let temp2 = Filename.temp_file "mlf" ".svg" in
  let r = Sys.command (Printf.sprintf "dot -Tsvg %s -o %s" temp1 temp2) in
  if r = 0 then (
    Sys.remove temp1;
    let ch = open_in temp2 in
    let s = Utils.read_file ~skip:8 ch in
    close_in ch;
    Sys.remove temp2;
    s
  ) else
    failwith "Dot error"
