open TypeConstructors
open SyntacticTypes
open SyntacticTypes.UnknownconstrTypes
open SyntacticTypesPrinting
open Locations
open Format


type value = [`Value] Opaque.string_t
let s_value s : value = Opaque.string_t s
let value_s (t : value) = Opaque.t_string t
module MapValues = Opaque.MapS(struct type t = [`Value] end)


(* This module defines the abstract syntax tree for the language
   of values and of toplevel expressions *)


(* Built-in constants corresponding to builtin constructions *)

type builtin_ct =
  | OpSeqNonUnit (* Sequence, type not constrained to unit *)
  | OpSeqUnit    (* Sequence, type     constrained to unit *)
  | OpIfThenElse (* If, then, else *)
  | OpIfThen     (* If, then, no else *)
  | OpFix (* Fixpoint operator *)
  | OpEq  (* Polymorphic equality *)
  | OpPlus (* plus on integers *)
  | OpMinus (* minus on integers *)

let print_builtin_ct ppf c = fprintf ppf "%s"
  (match c with
     | OpSeqNonUnit -> "#;"
     | OpSeqUnit -> "#();"
     | OpIfThenElse -> "#IfTE"
     | OpIfThen -> "#IfT"
     | OpFix -> "#fix"
     | OpEq -> "#eq"
     | OpPlus -> "#+"
     | OpMinus -> "#-"
  )


(* Patterns for variables bindings *)

type pattern =
  | PatVar of value
  | PatCoerce of pattern loc * expanded_polytype loc
(*
  | PUple of 'a pattern list * loc
  | PCons of string * 'a pattern * loc
  | PCtCons of string * loc *)

(* Simple form of patterns for functions *)
type unconstrained_let_name = [ `Unconstrained of value loc ]
type constrained_let_name = [
  | `Constrained of (value loc * expanded_polytype loc) loc
  | unconstrained_let_name
]


(* Annotations for recursive function *)
type nonrec = [ `NonRec ]
type rec_flag = [ nonrec | `Rec ]


(* The languages of values is split into the values that can only
   appear in the source language and the values that can only appear
   in the core language which is the basis for the constraint
   generator (plus a common part)
*)

(* Common values *)
type 'a expr_common_ = [
  | `Value of value
  | `Constr of constr_name
  | `Uple of 'c list
  | `Ct of value * init_ct_typeconstr monotype_
  | `App of 'c * 'c
  | `Abs of value loc * 'c
] constraint 'c = 'a loc

(* source only expressions *)
type 'a source_expr_ = [
  | 'a expr_common_
  | `Coerce of 'c * expanded_polytype loc
  | `SeqApp of 'c * 'c list
  | `SeqAbs of pattern loc list * 'c
  | `IfThenElse of 'c * 'c * 'c
  | `IfThen of 'c * 'c
  | `Equal of 'c * 'c
  | `Plus of 'c * 'c
  | `Minus of 'c * 'c
  | `Seq of 'c list * 'c
] constraint 'c = 'a loc
type 'a source_let = [
  `Let of rec_flag * (constrained_let_name  * 'a loc) * 'a loc
]

(* typing only expressions *)
type 'a typing_expr_ = [
  | 'a expr_common_
  | `CoerceCt of expanded_polytype
  | `BuiltinCt of builtin_ct
]
type 'a typing_let = [
  `Let of nonrec * (unconstrained_let_name  * 'a loc) * 'a loc
]

type source_expr = [
  | source_expr source_expr_
  | source_expr source_let
]
type typing_expr = [
  | typing_expr typing_expr_
  | typing_expr typing_let
]
(* The set of all values. Used only by the pretty-printer *)
type expr = [
  | expr source_expr_
  | expr typing_expr_
  | expr source_let
]


let builtin v = Locations.noloc (`BuiltinCt v)
let var_pat v = { v with obj = PatVar v.obj }

let let_dest v e1 e2 : typing_expr =
  `Let (`NonRec, (v, e1), e2)

let rec desugar : source_expr loc -> typing_expr loc = fun e_loc ->
  let e' = match e_loc.obj with
    (* Unchanged *)
    | `Value _ | `Constr _ | `Ct _ as v -> v

    (* Recursively desugared *)
    | `App (e1, e2) -> `App (desugar e1, desugar e2)
    | `Abs (v, e) -> `Abs (v, desugar e)
    | `Uple l -> `Uple (List.map desugar l)

    (* Translated to the application of a builtin constant *)
    | `Equal (e1, e2) -> (desugar_seq_app (builtin OpEq) [e1;e2]).obj
    | `Plus (e1, e2) -> (desugar_seq_app (builtin OpPlus) [e1;e2]).obj
    | `Minus (e1, e2) -> (desugar_seq_app (builtin OpMinus) [e1;e2]).obj
    | `IfThenElse (c, e1, e2) ->
        (desugar_seq_app (builtin OpIfThenElse) [c;e1; e2]).obj
    | `IfThen (c, e) ->
        (desugar_seq_app (builtin OpIfThen) [c; e]).obj

    (* Specific translation *)
    | `SeqAbs (l, e) -> (desugar_seq_abs l (desugar e)).obj
    | `Coerce (p, t) -> desugar_coerce p t
    | `SeqApp (e, l) -> (desugar_seq_app (desugar e) l).obj
    | `Seq (l, r) -> (desugar_seq r l).obj

    (* Let *)
    | `Let (rflag, (v, e1), e2) ->
        let e1' = match rflag, v with
          | `NonRec, `Unconstrained _ -> desugar e1
          | `NonRec, `Constrained {obj = v, t} ->
              {e1 with obj = desugar_coerce { e1 with loc = v.loc } t}
          | `Rec, `Unconstrained v ->
              let l = join v.loc e1.loc in
              let p = var_pat v in
              {obj = desugar_fix p e1; loc = l}
          | `Rec, `Constrained {obj = v, t; loc = locp} ->
              let l = join locp e1.loc in
              let p = {obj = PatCoerce (var_pat v, t); loc = locp} in
              {obj = desugar_fix p e1; loc = l}
        and v' = match v with
          | `Unconstrained v -> v
          | `Constrained { obj = v, _ } -> v
        in
        `Let (`NonRec, (`Unconstrained v', e1'), (desugar e2))

  in
  (* Desugaring is not supposed to change the locations. We
     restore a few locations, as some auxiliary functions can
     eat some parentheses or introduce ghost expressions *)
  { e_loc with obj = e' }

and desugar_fix p e =
  `App (builtin OpFix, desugar_seq_abs [p] (desugar e))

and desugar_coerce e t : typing_expr =
  `App ({ t with obj = `CoerceCt t.obj }, desugar e)

and desugar_pattern (pat : pattern loc) r =
  match pat.obj with
    | PatVar v -> { obj = v; loc = pat.loc }, r
    | PatCoerce (p, t) ->
        let (v, r) = desugar_pattern p r in
        v, { obj = let_dest (`Unconstrained v)
            { pat with obj = desugar_coerce { v with obj = `Value v.obj } t} r;
             loc = join pat.loc pat.loc (* TODO *) }

and desugar_seq_abs l r =
  let rec aux r = function
    | [] -> r
    | p :: q ->
        let v, t = desugar_pattern p r in
        let r = { obj = `Abs (v, t) ; loc = join p.loc r.loc } in
        aux r q
  in aux r (List.rev l)

and desugar_seq_app e l = desugar_seq_app_typing e (List.map desugar l)

and desugar_seq_app_typing e = function
  | [] -> e
  | e' :: q -> desugar_seq_app_typing { obj = `App (e, e');
                                        loc = join e.loc e'.loc } q
and desugar_seq r = function
  | [] -> desugar r
  | e :: q ->
      let seq = if !Options.sequence_has_type_unit
      then OpSeqUnit else OpSeqNonUnit in
      { (desugar_seq_app_typing (builtin seq) [desugar e; desugar_seq r q])
        with loc = join e.loc r.loc }


(* (Any) expression to string *)

let rec print_pattern ppf p =
  match p.obj with
    | PatVar v -> fprintf ppf "%s" (value_s v)
    | PatCoerce (p, t) -> fprintf ppf
        "(%a : %a)" print_pattern p print_type t.obj

let print_list_pattern = Utils.print_list print_pattern " "


module Level : sig
  type level
  val level_ct : level
  val level_app : level
  val level_equal : level
  val level_if : level
  val level_max : level

  val expr_level : expr -> level

  val levels_strictly_greater : level -> level -> bool
  val levels_equal : level -> level -> bool
end = struct
  type level = int

  let level_ct = 0
  let level_app = 5
  let level_equal = 10
  let level_plus_minus = 12
  let level_if = 15
  let level_seq = 20
  let level_max = 25

  let levels_strictly_greater = (>)
  let levels_equal = (=)

  let expr_level : expr -> level = function
    | `Value _ | `Constr _ | `Ct _ | `Uple _ |
          `Coerce _ | `CoerceCt _ | `BuiltinCt _ -> level_ct
    | `App _ | `SeqApp _ -> level_app
    | `Equal _ -> level_equal
    | `Plus _ | `Minus _ -> level_plus_minus
    | `IfThen _ | `IfThenElse _ -> level_if
    | `Seq _ -> level_seq
    | `Let _ | `Abs _ | `SeqAbs _ -> level_max
end
open Level


let debug_locations = ref false

let print_expr ppf (e : expr loc) =
  let mini = not !Options.minimize_parentheses in
  (* l is the level starting at which the expression will
     be parenthesize (if level e is strictly greater than l,
     or if level e = l and large is true) *)
  let rec expr ?(large=false) l () ppf e =
    let l' = expr_level e.obj in
    (* Shortcut function, which does the recursive call equal to
       the level of the current expression *)
    let expr ?(large=false) ?(level=l') () ppf e' =
      expr ~large level () ppf e'
    in
    if !debug_locations then fprintf ppf "@\n@[<hov 2>[%d-%d] "
      e.loc.loc_start.Lexing.pos_cnum e.loc.loc_end.Lexing.pos_cnum;
    let paren = levels_strictly_greater l' l || (large && levels_equal l' l) in
    if paren then fprintf ppf "(";
    (match e.obj with
      (* First level *)
      | `Value s | `Ct (s, _) -> fprintf ppf "%s" (value_s s)
      | `Constr c -> fprintf ppf "%s" (constr_s c)
      | `BuiltinCt c -> print_builtin_ct ppf c

      | `Uple l -> fprintf ppf "(%a)"
          (Utils.print_list (expr ~large:false ~level:level_app ()) ", ") l

      | `Coerce (e, t) -> fprintf ppf "(%a : %a)"
          (expr ()) e print_type t.obj

      | `CoerceCt t -> fprintf ppf "(: %a)" print_type t

      (* Application *)
      | `App (e1, e2) -> fprintf ppf "%a %a"
            (expr ~large:mini ()) e1 (expr ~large:true ()) e2
      | `SeqApp (e, l) -> fprintf ppf "%a %a"
            (expr ~large:mini ()) e
            (Utils.print_list (expr ~large:true ()) " ") l

      (* Operators *)
      | `Equal (e1, e2) ->
          fprintf ppf "%a = %a"
            (expr ~large:true ()) e1 (expr ~large:true ()) e2

      | `Plus (e1, e2) ->
          fprintf ppf "%a + %a"
            (expr ~large:false ()) e1 (expr ~large:false ()) e2
      | `Minus (e1, e2) ->
          fprintf ppf "%a - %a"
            (expr ~large:false ()) e1 (expr ~large:true ()) e2


      (* Conditionals *)
      | `IfThenElse (c, e1, e2) ->
          fprintf ppf "if %a then %a else %a"
            (expr ~large:mini ()) c
            (expr ~large:true ()) e1 (expr ~large:true ()) e2
      | `IfThen (c, e) ->
          fprintf ppf "if %a then %a"
            (expr ~large:mini ()) c (expr ~large:true ()) e

      (* sequence of expressions *)
      | `Seq (l, r) ->
          fprintf ppf "%a; %a"
            (Utils.print_list (expr ~large:true ()) "; ") l
            (expr ~large:mini ()) r

      (* Delimited high-level constructions *)
      | `Let (frec, (p, e1), e2) ->
          let v, t = match p with
            | `Constrained {obj = v, t} -> v, Some t
            | `Unconstrained v -> v, None
          in
          let_to_string ppf frec v t e1 e2

      | `Abs (v, e) -> fprintf ppf "fun %s -> %a"
            (value_s v.obj) (expr ()) e
      | `SeqAbs (l, e) -> fprintf ppf "fun %a -> %a"
            print_list_pattern l (expr ()) e
    );
    if paren then fprintf ppf ")";
    if !debug_locations then fprintf ppf "@]";


  and let_to_string ppf isrec v t e1 e2 =
      let r = if isrec = `Rec then " rec" else ""
      and v ppf t = match t with
        | None -> fprintf ppf "%s" (value_s v.obj)
        | Some t -> fprintf ppf "(%s : %a)"
            (value_s v.obj) print_type t.obj
      in
      match e1.obj with
        | `SeqAbs (l, e1) ->
            fprintf ppf "let%s %a %a = %a in %a"
              r v t print_list_pattern l
              (expr ~large:mini level_max ()) e1 (expr level_max ()) e2
        | _ ->
            fprintf ppf "let%s %a = %a in %a"
              r v t (expr ~large:mini level_max ()) e1 (expr level_max ()) e2
  in
  expr level_max () ppf e


let print_source_expr ppf e =
  print_expr ppf (e : source_expr loc :> expr loc)

let source_expr_to_string = Utils.format_into_string print_source_expr



(* ASTs for toplevel definitions *)

type type_decl = typeconstr_name loc * typevar list loc

type inductive_constr_decl = constr_name loc * expanded_polytype loc list

type inline =
  | InlineArrow of bool
  | InlineUple of sugar
  | InlineConstr of typeconstr_name * (sugar list) loc

type phrase =
  | Expr of source_expr loc * bool
  | Val of rec_flag * constrained_let_name * source_expr loc
  | Inductive of type_decl * inductive_constr_decl list loc
  | Alias of type_decl * expanded_polytype loc
  | Inline of inline
  | Include of string

let let_name : constrained_let_name -> _  = function
    `Unconstrained v | `Constrained { obj = v, _ } -> v

let val_to_let r name v : source_expr loc =
  let name' = let_name name in
  { name' with obj =
      `Let (r, (name, v), { name' with obj = `Value name'.obj })}
