{
  open Lexing
  open Parser
  open SyntaxErrors

  let keywords = [
    "fun", FUN;
    "let", LET;
    "in", IN;
    "if", IF;
    "then", THEN;
    "else", ELSE;
    "val", VAL;
    "type", TYPE;
    "of", OF;
(*    "match", MATCH;
    "with", WITH;*)
    "rec", REC;
  ]

  let directives = [
    "inline", INLINE_TYPES;
    "fail", FAIL;
    "include", INCLUDE;
  ]

  let newline lexbuf =
    lexbuf.lex_curr_p <- { lexbuf.lex_curr_p with
                             pos_lnum = lexbuf.lex_curr_p.pos_lnum + 1;
                             pos_bol =  lexbuf.lex_curr_p.pos_cnum;
                         }

  let in_comment = ref false
}

let newline = ('\n' | '\r' | "\r\n" "\n\r")
let any_ident_char = ['0'-'9'  'a'-'z'  'A'-'Z'  '_'  '\'']

rule token = parse
| [' ' '\t']+
    { token lexbuf }
| newline
    { newline lexbuf;
      token lexbuf }

| "(*"
    { in_comment := true;
      comment [Locations.curr_loc lexbuf] lexbuf;
      token lexbuf }

| ['a'-'z'] any_ident_char* as s
    { try List.assoc s keywords
      with Not_found -> LIDENT s }

| ['A'-'Z'] any_ident_char* as s
    { UIDENT s }

| '#'(['a'-'z' '_']+ as s)
    { try List.assoc s directives
      with Not_found ->
        raise (LexingError (InvalidDirective (Locations.curr_loc lexbuf, s)))
    }


| '\'' (['a'-'z']+ as s)
    { QUOTELIDENT s }

| ['0'-'9']+ as i
    { INT (int_of_string i) }

| '_'
    { UNDER }
| ':'
    { COLON }
| ','
    { COMMA }
| ';'
    { SEMI }
| ";;"
    { SEMISEMI }
| '.'
    { DOT }

| '>'
    { GREATER }
| '='
    { EQUAL }
| '*'
    { STAR }
(*| '\\'
    { FUN }*)
| '|'
    { PIPE }

| '+'
    { PLUS }
| '-'
    { MINUS }


| "->"
    { ARROW }

| '('
    { LPAREN }
| ')'
    { RPAREN }
| eof
    { EOF }
| _
    { raise (LexingError (InvalidChar (Locations.curr_loc lexbuf,
                                       Lexing.lexeme_char lexbuf 0))) }

and comment l = parse
| "*)"
    { match l with
        | [] -> assert false
        | [_] -> in_comment := false
        | _ :: q -> comment q lexbuf
    }
| "(*"
    { comment (Locations.curr_loc lexbuf :: l) lexbuf }
| newline
    { newline lexbuf;
      comment l lexbuf }
| eof { in_comment := false;
        raise (LexingError (UnterminatedComment (List.hd l))) }
| _
    { comment l lexbuf }
