%{
  open TypeConstructors
  open SyntacticTypes
  open UnknownconstrTypes
  open AST
  open Locations
  open SyntaxErrors


  let error r (loc : unit loc) =
    raise (ParsingError (ParseError (r, loc_ loc.loc.loc_start loc.loc.loc_end)))

  let unclosed symb pos1 pos1' decl symb' pos2 pos2' =
    raise (ParsingError (Unclosed
             ((symb, loc_ pos1 pos1'), decl, (symb', loc_ pos2 pos2'))))

  let tvar = let c = Utils.counter 0 in
  (fun () -> s_typevar (Printf.sprintf "#%d" c#get))


  let flex n ?(b=`Bottom) a : [> expanded_polytype] =
    `Bound (Flex, (n, b), a)
  let rigid n ?(b=`Bottom) a : [> expanded_polytype] =
    `Bound (Rigid, (n, b), a)

%}


%token EOF

%token <string> LIDENT UIDENT QUOTELIDENT
%token <int> INT
%token COMMA DOT COLON STAR SEMI SEMISEMI
%token GREATER EQUAL
%token ARROW FUN LET IN REC
%token LPAREN RPAREN
%token IF THEN ELSE
%token VAL
%token TYPE OF
%token PIPE
%token PLUS MINUS
%token UNDER
//%token MATCH WITH

(* Toplevel directives *)
%token INLINE_TYPES
%token FAIL
%token INCLUDE

%nonassoc THEN       (* Dangling else *)
%nonassoc ELSE       (* cf. THEN   *)

%right    ARROW      (* Associativity in type declaration *)

%nonassoc below_SEMI (* cf. seq_expr_aux *)

%nonassoc SEMI   (* below EQUAL for (fun x -> a ; (b=c)) vs. (fun x -> a ; b)=c
                    above IN for let x = v in (v ; v)  vs. (let x = v in v) ; v
                    above ARROW for fun p -> (v ; v)  vs. (fun p -> v) ; v *)

%left     EQUAL      (* cf. IN *)
%left PLUS MINUS


%start <SyntacticTypes.UnknownconstrTypes.expanded_polytype> polytype_ monotype_
%start <AST.source_expr Locations.loc> expr_
%start <AST.phrase Locations.loc list> phrases_
%start <AST.phrase> toplevel_phrase

%type <AST.source_expr> expr simple_expr
%type <SyntacticTypes.UnknownconstrTypes.expanded_polytype> simple_monotype monotype polytype

%%

%inline separated_list_mintwo(separator, X):
| x = X; separator; l = separated_nonempty_list(separator, X)
    { x :: l }

%inline paren(X):
| LPAREN; x = X; RPAREN
    { x }

%inline list_option(X):
| { [] }
| x = X  { x}

%inline l(X):
x = X { loc x $startpos(x) $endpos(x) }

%inline lerror:
e = l(error) { e }

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%% Types %%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%inline type_constructor:
| v = LIDENT
    { `PConstr (s_typeconstr v) }

%inline real_type_variable:
| v = QUOTELIDENT
    { s_typevar v }

%inline type_variable:
| v = real_type_variable
    { v }
| UNDER
    { tvar () }

simple_monotype:
| LPAREN m = polytype RPAREN
    { m }
| v = type_variable
    { `Var v }
| t = type_constructor
    { `Constr (t, []) }
(* Type constructor with a single parameter *)
| m = simple_monotype; t = type_constructor
    { `Constr (t, [m]) }
(* Type constructor with at least two parameters *)
|  args = paren(separated_list_mintwo (COMMA, polytype)); t = type_constructor
    { `Constr (t, args) }
//Errors
| LPAREN e = lerror
    { error "type or polytype bound expected (after '(' )" e }
| paren(separated_list_mintwo (COMMA, polytype)); e = lerror
    { error "type constructor expected after type constructor arguments" e }

monotype:
| m = simple_monotype { m }
(* type of an uple *)
| l = separated_list_mintwo(STAR, simple_monotype)
    { `Constr (`Uple (List.length l), l) }
(* Arrow type *)
| m1 = monotype ARROW m2 = monotype
    { `Constr (`Arrow, [m1; m2]) }
//Errors
| monotype ARROW e = lerror
    { error "type expected (after \"t -> \")" e }
| simple_monotype STAR e = lerror
    { error "type expected (after \"t *\")" e }

%inline sep_bound:
| GREATER
    { flex }
| EQUAL
    { rigid }

polytype:
| m = monotype
    { m }
| v = real_type_variable DOT p = polytype
    { flex v  p }
| LPAREN v = real_type_variable s = sep_bound b = polytype RPAREN DOT t = polytype
    { s v ~b:b t }
//Errors
| LPAREN real_type_variable e = lerror
    { error "'>' or '=' expected after variable name \
(in bound of quantified type)" e }
| real_type_variable DOT e = lerror
    { error "type expected after \"'a.\" (in bound of quantified type)" e }
| LPAREN real_type_variable sep_bound e = lerror
    { error "type expected after \"('a >\" or \"('a =\" \
(in bound of quantified type)" e }
| LPAREN real_type_variable sep_bound polytype RPAREN e = lerror
    { error "'.' expected after \"('a > t)\" \
(in quantified type)" e }
| LPAREN real_type_variable sep_bound polytype RPAREN DOT e = lerror
    { error "type expected after \"('a > t).\" \
(in quantified type)" e }


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%% Patterns %%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%inline pattern_var:
| v = LIDENT
    { s_value v }

pattern:
| v = pattern_var
    { PatVar v }
| LPAREN p = pattern RPAREN
    { p }
| LPAREN p = l(pattern) COLON t = l(polytype) RPAREN
    { PatCoerce (p, t) }
//Errors
| LPAREN e = lerror
    { error "pattern expected" e }
| LPAREN pattern e = lerror
    { error "\": t\" or ')' expected" e }
| LPAREN pattern COLON e = lerror
    { error "type expected after \"pat :\"" e }


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%% Values %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

simple_expr:
| i = INT
    { `Ct (s_value (string_of_int i), `Constr (`Int, [])) }
| v = LIDENT
    { (`Value (s_value v)) }
| c = UIDENT
    { `Constr (s_constr c) }
| LPAREN e = expr RPAREN
    { e }
| LPAREN RPAREN
    { `Ct (s_value "()", `Constr (`Unit, [])) }
| LPAREN e = l(expr) COLON t = l(polytype) RPAREN
    { `Coerce (e, t) }
| LPAREN l = uple_elements RPAREN
    { `Uple l }
| LPAREN l = seq_expr2 RPAREN
    { l }
//Errors
| LPAREN e = lerror
    { error "expression or ')' expected" e }
| LPAREN expr e = lerror
    { error "':', ',', ';' or ')' expected" e }
| LPAREN expr COLON e = lerror
    { error "type expected" e }

app_expr:
| e = l(simple_expr)
    { [e] }
| e = l(simple_expr)  q = app_expr
    { e :: q}

expr:
| e = simple_expr
    { e }
| FUN l = l(pattern)+ ARROW e = seq_expr
    { `SeqAbs (l, e) }
| e = l(simple_expr) l = app_expr (* l(simple_expr)+ *)
    { `SeqApp (e, l) }
| LET r = rec_flag v = constrained_let_name
    p = l(pattern)* EQUAL e1 = seq_expr IN e2 = seq_expr
    { let e1 =
        if p = [] then e1 else loc (`SeqAbs(p, e1)) $startpos(p) $endpos(e1)
      in `Let (r, (v, e1), e2) }
| e1 = l(expr) EQUAL e2 = l(expr)
    { `Equal (e1, e2) }
| e1 = l(expr) PLUS e2 = l(expr)
    { `Plus (e1, e2) }
| e1 = l(expr) MINUS e2 = l(expr)
    { `Minus (e1, e2) }
| IF c = seq_expr THEN e = l(expr)
    { `IfThen (c, e) }
| IF c = seq_expr THEN e1 = l(expr) ELSE e2 = l(expr)
    { `IfThenElse (c, e1, e2) }
//Errors
| FUN e = lerror
    { error "pattern expected after fun" e }
| FUN pattern+ e = lerror
    { error "pattern or '->' expected after \"fun ...\"" e }
| LET e = lerror
    { error "variable or 'rec' expected after \"let\"" e }
| LET rec_flag constrained_let_name e = lerror
    { error "pattern or '=' expected after \"let v\"" e }
| LET rec_flag constrained_let_name l(pattern)* EQUAL e = lerror
    { error "expression expected after \"let v =\"" e }
| LET rec_flag constrained_let_name l(pattern)* EQUAL seq_expr e = lerror
    { error "'in' expected after \"let v = ...\"" e }
| LET rec_flag constrained_let_name l(pattern)* EQUAL seq_expr IN e = lerror
    { error "expression expected after \"let ... in\"" e }
| IF e = lerror
    { error "expression expected after \"if\"" e }
| IF seq_expr THEN e = lerror
    { error "expression expected after \"if ... then\"" e }

constrained_let_name:
| v = l(pattern_var)
    { `Unconstrained v }
| LPAREN v = l(pattern_var) COLON t = l(polytype) RPAREN
    { `Constrained (loc (v, t) $startpos $endpos) }

seq_expr_aux:
| e = l(expr)  %prec below_SEMI (* fun p -> (v1 ; v2)  vs . (fun p -> v1) ; v2
                                  with the first v1 going through seq_expr *)
    { [], e }
| e = l(expr) SEMI l = seq_expr_aux
    { let l, e' = l in e :: l, e' }
//errors
//| expr e = lerror
//    { error "';' expected (in sequence of expressions)" e }
| expr SEMI e = lerror
    { error "'expression expected after ';' (in sequence of expressions)" e }

%inline seq_expr: r = seq_expr_aux
    { let l, e = r in
      if l = [] then e else loc (`Seq (l, e)) $startpos $endpos }

(* Sequences with at least two elements. Avoids a conflict
   because LPAREN expr RPAREN -> expr is valid (through simple_expr)
   with and without sequences *)
%inline seq_expr2:
| e = l(expr) SEMI l = seq_expr_aux
    { let l, e' = l in `Seq (e :: l, e') }
| expr SEMI e = lerror
    { error "'expression expected after ';' (in sequence of expressions)" e }


(* For uples *)
uple_elements:
| e1 = l(expr) COMMA e2 = l(expr)
    { [e1;e2] }
| e = l(expr) COMMA l = uple_elements
    { e :: l }
//errors
| expr COMMA expr e = lerror
    { error "',' or ')' expected (in uple)" e }
| expr COMMA e = lerror
    { error "expression expected (in uple)" e }

rec_flag:
|     { `NonRec }
| REC { `Rec }


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%% Type declaration %%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%inline constr_name:
| c = UIDENT
    { s_constr c }

constructor_decl:
| c = l(constr_name)
    { (c, []) }
| c = l(constr_name) OF args = separated_nonempty_list(STAR, l(simple_monotype))
    { (c, args) }
//errors
| UIDENT OF e = lerror
    { error "'*' separated sequence of simple types expected after 'of'" e }

constructors_decls:
| PIPE? l = l(separated_nonempty_list(PIPE, constructor_decl))
    { l }
//errors
| PIPE e = lerror
    { error "constructor declaration expected after '|'" e }

type_parameters:
| v = real_type_variable
    { [v] }
| LPAREN l = separated_nonempty_list(COMMA, real_type_variable) RPAREN
    { l }
//errors
| LPAREN e = lerror
    { error "type variable expected (in type parameters)" e }

%inline type_name:
| t = LIDENT
    { s_typeconstr t }

type_decl_prefix:
| TYPE p = l(list_option(type_parameters)) t = l(type_name) EQUAL
    { t, p }
//errors
| TYPE e = lerror
    { error "type name or type parameters expected after \"type\"" e }
| TYPE type_parameters e = lerror
    { error "type name expected after type parameters" e }
| TYPE list_option(type_parameters) type_name e = lerror
    { error "'=' expected after type name (in type declaration)" e }
| type_decl_prefix e = lerror
    { error "type or '|' or constructor declaration expected" e }


inductive_declaration:
| p = type_decl_prefix l = constructors_decls
    { p, l }

alias_declaration:
| p = type_decl_prefix t = l(polytype)
    { p, t }

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%% Module phrases %%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%inline phrase_expr:
| e = seq_expr
    { Expr (e, true) }

%inline phrase_let:
| VAL r = rec_flag v = constrained_let_name l = l(pattern)* EQUAL e = seq_expr
    { let e = if l = [] then e
      else loc (`SeqAbs (l, e)) $startpos(l) $endpos(e) in
      Val (r, v, e) }
//Errors
| VAL e = lerror
    { error "variable or 'rec' expected after \"val\"" e }
| VAL rec_flag LIDENT e = lerror
    { error "variable or '=' expected after \"val ...\"" e }

%inline phrase_type:
| i = inductive_declaration
    { let t, c = i in Inductive (t, c) }
| a = alias_declaration
    { let t, t' = a in Alias (t, t') }

%inline sugar:
| MINUS { None }
| GREATER { Some Flex }
| EQUAL { Some Rigid }

%inline phrase_directive:
| INLINE_TYPES ARROW i = INT
    { Inline (InlineArrow (i > 0)) }
| INLINE_TYPES LPAREN DOT RPAREN s = sugar
    { Inline (InlineUple s) }
| INLINE_TYPES i = LIDENT l = l(list(sugar))
    { Inline (InlineConstr (s_typeconstr i, l)) }
| FAIL e = seq_expr
    { Expr (e, false) }
| INCLUDE f = UIDENT
    { Include f }
//errors
| INLINE_TYPES e = lerror
    { error "'->', '(.)' or constructor expected after #inline" e }
| INLINE_TYPES ARROW e = lerror
    { error "0 or 1 expected after #inline ->" e }
| INLINE_TYPES LPAREN DOT RPAREN e = lerror
    { error "sugar annotation expected after #inline (.)" e }
| FAIL e = lerror
    { error "expression expected after #fail" e }
| INCLUDE e = lerror
    { error "Uppercase file name expected after #include" e }

phrase_middle:
| SEMISEMI  e = phrase_expr
| SEMISEMI? e = phrase_let
| SEMISEMI? e = phrase_type
| SEMISEMI? e = phrase_directive
    { e }
//errors
| e = lerror
    { error "'val', 'type', toplevel directive or end of expression expected" e }
| SEMISEMI e = lerror
    { error "expression, 'val', 'type' or toplevel directive expected" e }

phrase:
| e = phrase_expr
| e = phrase_let
| e = phrase_type
| e = phrase_directive
    { e }

phrases:
| p = l(phrase)
    { [p] }
| l = phrases p = l(phrase_middle)
    { p :: l }
//errors
| e = lerror
    { error "expression, toplevel definition or toplevel directive expected" e }


toplevel_phrase:
| e = phrase SEMISEMI {e}
| EOF { raise End_of_file }

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%% Entry points %%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

expr_: e = seq_expr EOF { e }
| e = lerror { error "beginning of expression expected" e }
monotype_: t = monotype  EOF { t }
polytype_: t = polytype  EOF { t }
phrases_ :
| l = phrases SEMISEMI? EOF { List.rev l }
| EOF { [] }
