(*pp camlp4o ./hashtbl_ext.cmo *)

open TypeConstructors
open SyntacticTypes
open GraphicTypes
open GraphicTypes.Nodes
open GraphicTypesAux
open ConstraintsMeaning
open Utils
open Locations
open Highlight
open Format


type constraint_type = ML | MLF


let new_forall () =
  Forall ({ forall = 0 })

let forall_count f = match !!f.constr with
  | Forall d -> d
  | _ -> assert false

let incr_forall_count f =
  let d = forall_count f in
  d.forall <- d.forall + 1

let decr_forall_count f =
  let d = forall_count f in
  d.forall <- d.forall - 1


let bind_flexibly_to ?binder ~n =
  !!n.bound <- match binder with
    | None -> Root
    | Some n' -> Bound ({
        label = Flex;
        binder = n';
        permissions = PFlex;
        binding_height = binding_height n'+1;
      })

type constraint_edge_type = Unif | Inst
type constraint_edge = {
  edge_type : constraint_edge_type;
  edge_meaning : constraint_meaning;
  edge_nodes : node * node;
}
type constraints = {
  root: node;
  constraints: constraint_edge store;
  exists: node store;
}

let solved c =
  c.constraints#get = [] && c.exists#get = []
;;

let equal_pair_nodes (n1, n2) (n1', n2') =
  Nodes.equal n1 n1' && Nodes.equal n2 n2'

(* By sort well-formedness, there cannot both an unif edge and instantiation
   edge between the same nodes *)
let equal_constraints c1 c2  =
  equal_pair_nodes c1.edge_nodes c2.edge_nodes
let equal_constraints' p c  =
  equal_pair_nodes c.edge_nodes p


let binder_constraint c n =
  match !!n.bound with
    | Root -> c.root
    | Bound { binder = n' } -> n'

let existentials_on c n =
  List.filter (fun n' -> Nodes.equal n (binder_constraint c n'))
    c.exists#get


let constraints_succs c n = !!n.children @ existentials_on c n
let dfs_constraints c = GraphicTypesAux.aux_dfs (constraints_succs c)


(* Raising of nodes through unification. *)
let raise_to unif n b =
  let n' = new_node Bot [] in
  bind_flexibly_to ~binder:b ~n:n';
  unif (n, n');
;;

let raise_once c unif n =
  match !!n.bound with
    | Root -> false
    | Bound { binder = n' } ->
        let binder = match !!n'.bound with
          | Root -> c.root
          | Bound { binder = b } -> b
        in
        raise_to unif n binder
;;


let () = #{HashNodes}

(* Copy the constraints incoming on some copied nodes. *)
let constraints_copy c dest =
  let copied n = Nodes.equal (dest n) n = false in
  let rec copy = function
    | [] -> []
    | { edge_nodes = (n1, n2) } as e :: q ->
        if copied n1 || copied n2 then (
          (* We must update the forall counts on nodes that are
             outside of the copy *)
          let n1' = dest n1 in
          (if Nodes.equal n1 n1' = true then
             match !!n1'.constr with
               | Forall d -> d.forall <- d.forall + 1
               | _ -> ());
          e :: { e with edge_nodes = (n1', dest n2) } :: copy q;
        )
        else
          e :: copy q
  in
  c.constraints#set (copy c.constraints#get);
;;

let dfs_interior succs ?(frontier=unitf) ?(pre=unitf) ?(post=unitf) n_ =
  let visited = {} in
  let rec aux n =
    if visited{?n} = false then (
      visited{n} <- ();
      match !!n.bound with
        | Bound {binder = n'} when
            Nodes.equal n n_ = false && visited{?n'} = false ->
            (* In the frontier, as binder(n) is not in the interior *)
            frontier n;
        | _ ->
            pre n;
            List.iter aux (succs n);
            post n;
    )
    in aux n_


(* Perform expansion on a forall *)
let expand_forall c ?(constraint_type=MLF) ?(solve_easy_frontier=true) e =
  let (f, d) = e.edge_nodes
  and m = e.edge_meaning in
  let f' = child1 f in
  let copies = {} in
  let dest n = try copies{n} with Not_found -> n in
  let bd = binder d
  and bf = binder f in
  let lca =
    if Nodes.equal bf bd then
      bf
    else
      lca_binders (SetNodesHeight.union
         (SetNodesHeight.singleton bf) (SetNodesHeight.singleton bd))
  in
  let hlca = binding_height lca in

  let pre_mlf n =
    if Nodes.equal n f = false then
      match !!n.bound with
        | Root -> assert false
        | Bound { binder = n'; label = l } ->
            (* we use an intermediate node
               on which the nodes under can be bound *)
            let c_n = new_node Bot [] in
            (* Permissions reset *)
            if Nodes.equal f' n then
              bind_flexibly_to ~binder:(copies{n'}) ~n:c_n
            (* Binding reset *)
            else if Nodes.equal n' f then
              bind_to ~binder:(copies{f'}) ~n:c_n l
            (* General case *)
            else
              bind_to ~binder:(copies{n'}) ~n:c_n l;
            copies{n} <- c_n;

  and pre_ml n =
    if Nodes.equal n f = false then
      (match !!n.bound with
         | Root -> assert false
         | Bound { binder = bn; label = l} ->
             if l <> Flex || Nodes.equal bn f = false then
               raise TypingErrors.MLFTypeMLFExpansion);
      let c_n = new_node Bot [] in
      bind_flexibly_to ~binder:bd ~n:c_n;
      copies{n} <- c_n;

  and post n =
    if Nodes.equal n f = false then
      let children' = List.map dest !!n.children in
      let c_n = new_node !!n.constr children' in
      let c_n' = copies{n} in
      fuse ~keep:c_n ~bound:c_n' ~lose:c_n' ()

  and frontier n =
    if solve_easy_frontier && binding_height n >= hlca then (
      copies{n} <- n)
    else (
      let n' = new_node Bot [] in
      bind_flexibly_to ~binder:(if solve_easy_frontier then lca else bd) ~n:n';
      c.constraints#add ({ edge_type = Unif; edge_meaning = CRaise m;
                           edge_nodes = (n', n) });
      copies{n} <- n'
    )

  in
  let pre = match constraint_type with MLF -> pre_mlf | ML -> pre_ml in
  (* We copy all nodes of the interior of f, except f itself
     (pre and post do nothing on it *)
  copies{f} <- binder d;
  dfs_interior children ~pre ~post ~frontier f;

  let cf' = dest f' in
  c.exists#add cf';
  cf';
;;


(* Perform a reset on a type scheme : all the nodes bound on the forall
   are lowered, and the permissions/binding height are recomputed *)
let reset_forall f =
  let f' = child1 f in
  if Nodes.equal f (binder f') then (
    bind_flexibly_to ~n:f' ~binder:f;
    let aux n =
      if Nodes.equal n f = false && Nodes.equal n f' = false then
        if Nodes.equal (binder n) f then
          bind_to ~binder:f' ~n (blabel n)
        else
          recompute_node_infos n
    in
    dfs_interior children ~pre:aux f)


(* Monormophic nodes with the pro viso that existential nodes by
   are not monomorphic. Currently for display purposes *)
let is_mono_constraint c =
  let is_mono = GraphicTypesAux.mono ~dfs:(dfs_constraints c) c.root in
  (fun n -> is_mono n (*&& c.exists#mem n = false*))


open GraphicSyntacticTranslations

let forall_printing ?(nf=(!Options.normal_forms)) c =
  assert (solved c);
  let r = c.root in
  let r' = child1 r in
  let tr = graph_to_syntactic_aux ~nf () in
  let t = tr.node_type r' in
  rebind_context t (tr.node_context c.root)
;;
