open TypeConstructors
open Constraints
open ConstraintsMeaning
open SyntacticTypes
open KnownconstrTypes
open GraphicTypes
open GraphicTypes.Nodes
open AST
open Utils
open Locations
open TypingEnv
open TypingErrors


let builtin_ct_type =
  let t_sequnit = TypingEnv.initial_ct "'a. unit -> 'a -> 'a"
  and t_seqnonunit = TypingEnv.initial_ct "'a. 'b. 'b -> 'a -> 'a"
  and t_ite = TypingEnv.initial_ct "'a. bool -> 'a -> 'a -> 'a"
  and t_it = TypingEnv.initial_ct "bool -> unit -> unit"
  and t_fix = TypingEnv.initial_ct "'a. ('a -> 'a) -> 'a"
  and t_eq = TypingEnv.initial_ct "'a. 'a -> 'a -> bool"
  and t_plus_minus = TypingEnv.initial_ct "int -> int -> int"
  in function
    | OpSeqUnit -> t_sequnit
    | OpSeqNonUnit -> t_seqnonunit
    | OpIfThenElse -> t_ite
    | OpIfThen -> t_it
    | OpFix -> t_fix
    | OpEq -> t_eq
    | OpPlus | OpMinus -> t_plus_minus


(* Used on type variables *)
let () = #{Hashtbl}

(* A structure holding a link between existentially quantified
   variables inside coercions. Each type variable is mapped to the
   list of its occurrences in the constraint *)
type existential_variables =
    (typevar, (node * location) list) Hashtbl.t

(* Adding an occurrence of an existential variable *)
let add_ex_var (ex_vars : existential_variables) v n =
  let lnodes_v = try ex_vars{v} with Not_found -> [] in
  let lnodes_v' = n :: lnodes_v in
  ex_vars{v} <- lnodes_v'

(* Constraining all the occurrences of an existential variable
   to be equal *)
let constrain_existential_variables (ex_vars : existential_variables) (lc : constraint_edge store) =
  let constrain_existential_variable v = function
    | [] -> assert false
    | (n1, _) :: q ->
        List.iter (fun (n2, loc) ->
                     let e = { edge_nodes = (n1, n2); edge_type = Unif;
                               edge_meaning = CExistsVar (v, loc);
                             } in
                     lc#add e) q
  in
  Hashtbl.iter (fun v l -> constrain_existential_variable v l) ex_vars


let translate_known_type binder (t : polytype) =
  let s = new_node (Type t) [] in
  bind_flexibly_to ~binder ~n:s;
  if !Options.expand_types_from_env then Unif.expand_type s;
  s


let translate_coercion binder ex_vars (t : polytype) loc =
  (* height is the binding height of the root of the translation of the
     coercion corresponding to t *)
  let height = binding_height binder + 1
  in
  (* Free type variables of the coercion *)
  let ftv = ftv t
  (* Env is an environment binding the free (existentials) variables
     of t to fresh nodes, which will be used during the translation *)
  and env = {}
  in
  (* ftv' is a list of the nodes corresponding to the ftv, which will
     be rebound once the root of the coercion graph is known *)
  let ftv' = SetTypeVars.fold
    (fun v l ->
       let nv = new_node Bot [] in
       (* We add v to the existential vars of the constraint *)
       add_ex_var ex_vars v (nv, loc);
       (* And we update env and ftv' *)
       env{v} <- nv;
       nv :: l)
    ftv []
  in
  (* The type corresponding to the translation. The free variables are
     intentionnaly not bound *)
  let t_free =
    rigid (s_typevar "__a") ~b:t
      ((if !Options.rigid_right_coercions then rigid else flex)
         (s_typevar "__b") ~b:t (arrow (var "__a") (var "__b")))
  in
  let s = GraphicSyntacticTranslations.polytype_to_graph
    ~env ~perms:PFlex ~bheight:(height+1) t_free
  in
  (* We bind correctly the root of the coercion graph *)
  bind_flexibly_to ~binder ~n:s;
  (* And the free variables *)
  List.iter (fun n -> bind_flexibly_to ~binder:s ~n) ftv';
  s



(* Meaning of constraint edges in an application. We intercept
   constructs resulting from desugaring *)
let meaning_app e e1 e2 =
  let default = CApp (e1.loc, e2.loc) in
  match e1.obj with
    | `CoerceCt _ -> CAppCoercion (e1.loc, e2.loc)
    | `BuiltinCt c ->
        (* All builtin cts must be caught *)
        (match c with
           | OpSeqUnit -> CSeqUnit e2.loc
           | OpSeqNonUnit -> CSeqNonUnit e2.loc
           | OpIfThenElse | OpIfThen -> CIfCond e2.loc
           | OpFix -> CFix e2.loc
           | OpPlus | OpMinus -> CInt e2.loc
           | OpEq -> CEqFirst e2.loc
        )
    | `App ({ obj = `BuiltinCt c}, e1') ->
        (* All constants of arity at least 2 must be caught *)
        (match c with
           | OpIfThen -> CIfMissingElse (e2.loc, e.loc)
           | OpIfThenElse -> CIfThenWithElse (e2.loc, e.loc)
           | OpEq -> CEq (e1'.loc, e2.loc)
           | OpPlus | OpMinus -> CInt e2.loc
           | OpSeqUnit | OpSeqNonUnit -> CEndSeq e2.loc
           | OpFix -> default)
    | `App ({obj = `App ({obj = `BuiltinCt OpIfThenElse}, _)}, e1') ->
        CIfEq (e1'.loc, e2.loc, e.loc)
    | _ -> default


(* Search in the typing environment *)
let find_value loc env v =
  try MapValues.find v env.values
  with Not_found -> typing_error (UnboundVar (v, loc))

let find_constr loc env c =
  try fst (MapConstrs.find c env.constructors)
  with Not_found -> typing_error (UnknownInductiveConstructor (c, loc))


let inst_elim_mono e =
  match e.edge_type with
    | Unif -> e
    | Inst ->
        let f, d = e.edge_nodes in
        let n = child1 f in
        if Nodes.equal (binder n) f then e
        else (
          decr_forall_count f;
          { edge_type = Unif;
            edge_meaning = e.edge_meaning;
            edge_nodes = n, d }
        )

let translate ?(constraint_type = MLF) env (e : typing_expr loc) =
  let constraints = new store ~eq:equal_constraints ()
  and l_exists = new store ~eq:Nodes.equal ()
  and l_expr_roots = new store ()
  (* Existential variables in annotations *)
  and ex_vars : existential_variables = {}
  in
  let exists n = l_exists#add n
  (* Add a constraint between two nodes. Find the right kind
     of constraint by finding whether the first node is a forall *)
  and constrain ((n, _) as c) meaning =
    match !!n.constr with
       | Forall _ ->
           incr_forall_count n;
           constraints#add ({ edge_type = Inst; edge_meaning = meaning;
                              edge_nodes = c })
       | _ ->
           if Nodes.equal (fst c) (snd c) = false then
             constraints#add ({ edge_type = Unif; edge_meaning = meaning;
                                edge_nodes = c })
  in
  (* Translation of an expression into a constraint. Optimisations
     for Var-*, known constants, ML constraints and successive abstractions
     are done on the fly *)
  (* translates an expression inside a given breakpoint.
     f is the current forall for the shape of the constraint.
     fbind is the forall to which to bind nodes *)
  let rec translate vars fbind f e =
    let loc = e.loc in

    let new_bound_node c l =
      let n = new_node c l in
      bind_flexibly_to ~binder:fbind ~n;
      n
    in

    let n = match e.obj with
      | `Value v ->
          (try
             let { obj = n'; loc = n'_loc } = MapValues.find v vars in
             let bot = new_bound_node Bot [] in
             constrain (n', bot) (CVar (e.loc, n'_loc));
             bot
           with Not_found ->
             let t = find_value loc env v in
             translate_known_type fbind t
          )

      | `Constr c ->
          let t = find_constr loc env c in
          translate_known_type fbind t

      | `BuiltinCt c ->
          translate_known_type fbind (builtin_ct_type c)

      | `App (e1, e2) ->
          let bota = new_bound_node Bot []
          and botr = new_bound_node Bot [] in
          let arr = new_bound_node (Constr `Arrow) [bota; botr] in
          exists arr;

          let ffun = create_forall vars fbind f e1 in
          let farg = create_forall vars fbind f e2 in

          let meaning = meaning_app e e1 e2 in
          let c_fun () = constrain (ffun, arr) meaning
          and c_arg () = constrain (farg, bota) meaning in
          if !Options.solve_app_before_arg
          then (c_fun (); c_arg ())
          else (c_arg (); c_fun ());
          botr

      | `Abs (v, p) ->
          let bota = new_bound_node Bot []
          and botr = new_bound_node Bot [] in
          let arr = new_bound_node (Constr `Arrow) [bota; botr] in

(*        match !Options.shortcut_seqabs, prev_abs with
              | true, Some n ->
                  let arg, res = translate_abs_shortcut n in n, arg, n, res
              | _ -> translate_abs ?binder () *)

          let vars' = MapValues.add v.obj ({ v with obj = bota }) vars in
          (match p.obj, !Options.shortcut_seqabs with
            | `Abs _, true ->
                let fbind' =
                  match constraint_type with ML -> fbind | MLF -> botr in
                let n = translate vars' fbind' botr p in
                fuse ~keep:n ~lose:botr () ~bound:botr;
                arr
            | _ ->
                let fp = create_forall vars' fbind f p in
                constrain (fp, botr) (CAbs loc);
                l_expr_roots#add (bota, v.loc);
                arr
          )

      (* let x = foo in x  -->  foo *)
      | `Let (`NonRec, (`Unconstrained v, e), e') when e'.obj = `Value v.obj ->
          translate vars fbind f e


      | `Let (`NonRec, (`Unconstrained v, e1), e2) ->
          (* TODO : in ML we can choose between f and fbind as the binder
             of f1. The correct answer is f, but fbind should make
             for more efficient constraints *)
          let f1 = create_real_forall vars ~f:f e1 in
          let vars' = MapValues.add v.obj ({ v with obj = f1 }) vars in
          let f2 = translate vars' fbind f e2 in
          l_expr_roots#add (f1, v.loc);
          f2

      | `Ct (_, t) ->
          translate_known_type fbind
            (t : init_ct_typeconstr monotype_ :> typeconstr polytype_)

      | `CoerceCt t ->
          let t' = unexpand_resolve_constrs loc env t in
          (if constraint_type = ML then match t' with
             | #monotype -> ()
             | #polytype ->
                 typing_error (MLFCoercionMLConstraint (t, loc)));
          translate_coercion fbind ex_vars t' loc

      | `Uple l ->
          let bots = List.map (fun _ -> new_bound_node Bot []) l in
          let u = new_bound_node (Constr (`Uple (List.length l))) bots in
          List.iter (fun n -> bind_flexibly_to ~binder:u ~n) bots;
          List.iter2 (fun e bot ->
                        let f' = create_forall vars fbind f e in
                        constrain (f', bot) (CUple (e.loc, loc));
                     ) l bots;
          u

    in l_expr_roots#add (n, loc); n

  (* Creates a possibly degenerate generalization scope. In ML, subexpressions
     are bound to fbind. In MLF most expressions are bound to f, except
     those for which the case is provably degenerate *)
  and create_forall vars fbind f e =
    let var : [ `LocalVar of value | `EnvVar | `NonVar ] = match e.obj with
      | `Value v ->
          if MapValues.mem v vars then `LocalVar v
          else `EnvVar
      | _ -> `NonVar
    in
    match e.obj, var, !Options.shortcut_bottoms, !Options.shortcut_forall with
      | _, `LocalVar v, true, _ ->
          (* Generalization scopes for local variables can be removed
             using Var-* *)
          (MapValues.find v vars).obj

      | (`CoerceCt _ | `Ct _ | `BuiltinCt _ | `Constr _), _, _, true
      | _, `EnvVar, _, true ->
          (* Optimized case; the forall is useless, we do not create it
             (thus an unification edge will be used instead) *)
          let n = translate vars fbind f e in
          exists n; n

      | _ ->
          (* Default case, we create the forall *)
          let b = new_node Bot [] in
          let f' = new_node (Forall ({forall = 0})) [b] in
          exists f';
          bind_flexibly_to ~binder:f ~n:f';
          let binder = match constraint_type with MLF -> f' | ML -> fbind in
          let r = translate vars binder f' e in
          fuse ~keep:r ~lose:b ();
          f'


(* Creates a real generalization point for the left hand side of
   a let, or the root of the constraint *)
  and create_real_forall vars ?f e =
    let b = new_node Bot [] in
    let f' = new_node (Forall ({forall = 0})) [b] in
    bind_flexibly_to ?binder:f ~n:f';
    if f <> None then exists f';
    let n = translate vars f' f' e in
    fuse ~keep:n ~lose:b ();
    f'

  in
  let f = create_real_forall MapValues.empty e in
  (* Constraints are generated in the reverse order needed for
     automatic resolution *)
  constraints#set (List.rev constraints#get);
  constraints#set (List.map inst_elim_mono constraints#get);
  constrain_existential_variables ex_vars constraints ;

  { root = f; constraints = constraints; exists = l_exists },
  l_expr_roots

let translate_initial_env =
  translate TypingEnv.initial_env
