open GraphicTypes
open Nodes
open Locations

type constraint_meaning =
(* Constraints that should never fail *)
  (* Return type of an abstraction. *)
  | CAbs of location

  (* First element of a comparison *)
  | CEqFirst of location

  (* Beginning of a sequence which needs not have type unit *)
  | CSeqNonUnit of location

  (* Then part of an if with an else (hence unconstrained).
     loc of then, loc of whole if *)
  | CIfThenWithElse of location * location

  (* End of a sequence *)
  | CEndSeq of location

  (* Nodes to be raised by an expansion *)
  | CRaise of constraint_meaning


(* Constraints that should not fail, *given the current resolution
   strategy *. Hence their error messages need not be very precise *)
  (* Element of an uple, whole uple. *)
  | CUple of location * location

  (* An existential variable inside an annotation (whole annotation) *)
  | CExistsVar of TypeConstructors.typevar * location


(* Constraints that can fail *)
  (* The occurrence of the variable, where the variable is bound *)
  | CVar of location * location

  (* The function to be applied, its argument *)
  | CApp of location * location

  (* A coercion applied to an expression *)
  | CAppCoercion of location * location

  (* The condition of a if. Must have type bool *)
  | CIfCond of location

  (* A 'if' without a 'else' ; must have type unit.
     location of the then part, location of the entire if *)
  | CIfMissingElse of location * location

  (* Both branches of a 'if' must have the same type *)
  (* location of the 'then', of the 'else', of the whole 'if' *)
  | CIfEq of location * location * location

  (* A use of the '=' operator. Both sides must have equal types *)
  | CEq of location * location

  (* A statement inside a sequence. Must have type unit *)
  | CSeqUnit of location

  (* Fixpoint operator *)
  | CFix of location

  (* Arguments of plus or minus *)
  | CInt of location

  (* Constraint not generated, but used to display the fact that an
     expression is applied but is not a function *)
  | CNotFun of location



let fragment_from_loc code loc =
  Printf.sprintf "\"%s\""
    (String.sub code loc.loc_start.Lexing.pos_cnum
       (loc.loc_end.Lexing.pos_cnum - loc.loc_start.Lexing.pos_cnum))

let fragment_from_meaning code m =
  let loc = fragment_from_loc code in
  match m with
    | CNotFun _ -> assert false

    | CRaise _ -> "Raising a frontier node"

    | CApp (l1, l2) -> Printf.sprintf "%s applied to %s" (loc l2) (loc l1)

    | CAppCoercion (loc_t, loc_e) ->
        Printf.sprintf "%s constrained to %s" (loc loc_e) (loc loc_t)

    | CAbs l ->
        Printf.sprintf "Return type of %s" (loc l)

    | CVar (l_occ, _l_bound) ->
        Printf.sprintf "Variable %s" (loc l_occ)

    | CUple (loc_e, loc_t) ->
        Printf.sprintf "Element %s of the uple %s" (loc loc_e) (loc loc_t)

    | CFix l ->
        Printf.sprintf "Recursive value %s" (loc l)

    | CEqFirst l ->
        Printf.sprintf "First element %s of a comparison through '='" (loc l)

    | CEq (l1, l2) ->
        Printf.sprintf "%s and %s compared by '='" (loc l1) (loc l2)

    | CSeqNonUnit l ->
        Printf.sprintf "%s as beginning of a sequence" (loc l)

    | CSeqUnit l ->
        Printf.sprintf "%s as beginning of a sequence, constrained to unit" (loc l)

    | CIfCond l ->
        Printf.sprintf "Comparison %s in an 'if'" (loc l)

    | CInt l ->
        Printf.sprintf "%s applied to an arithmetic operator" (loc l)

    | CIfEq (_lthen, lelse, _lif) ->
        Printf.sprintf "'Else' part %s of an if" (loc lelse)

    | CIfMissingElse (lthen, _lif) ->
        Printf.sprintf "'then' part %s of an 'if' without 'else'" (loc lthen)

    | CIfThenWithElse (lthen, _lif) ->
        Printf.sprintf "%s as 'then'" (loc lthen)

    | CEndSeq l ->
        Printf.sprintf "%s as end of sequence" (loc l)

    | CExistsVar (v, _l) ->
        Printf.sprintf "existential type variable %s"
          (TypeConstructors.typevar_s v)
