open GraphicTypes
open Nodes
open Constraints
open ConstraintsMeaning

(* Exceptions resulting from the solving of constraints *)
exception InferenceError of Unif.unif_error * constraint_meaning


let existentials_from_edges c =
  let st = new Utils.store ~eq:Nodes.equal () in
  List.iter (fun { edge_nodes = (n1, n2) } -> st#add n1; st#add n2)
    c.constraints#get;
  st

let check_without_existentials c =
  assert (Wf.correct ({c with exists = existentials_from_edges c}))

let no_check (_ : constraints) = ()

(* Fast solving of the first constraint, with error interception at the end *)
let fast_solve_one env c e =
  let n1, n2 = e.edge_nodes in
  let n1', n2' = match e.edge_type with
    | Unif -> (n1, n2)
    | Inst ->
        decr_forall_count n1;
        expand_forall c e, n2
  in
(* Catch an unexplained unif error and adds the meaning of the constraint
   whose resolution failed. *)
  try Unif.unif env (n1', n2')
  with Unif.UnifError error ->
    raise (InferenceError (error, e.edge_meaning))

(* Optimisation : we solve all unification edges first, as at the beginning
   of the constraint they usually only involve bottom nodes
   (solving them later which would require updating binders *)
let solve_all_unif ~env c =
  let rec aux = function
    | [] -> []
    | { edge_type = Unif } as e :: q ->
        fast_solve_one env c e;
        aux q
    | { edge_type = Inst } as e :: q ->
        e :: aux q
  in
  c.constraints#set (aux c.constraints#get)

let solve ~env ?(check=check_without_existentials) c =
  let aux () = match c.constraints#get with
    | [] -> true
    | e :: q ->
        c.constraints#set q;
        fast_solve_one env c e;
        false
  in
  c.exists#set [];
  solve_all_unif ~env c;
  while aux () = false do
    c.exists#set [];
    check c;
  done;
;;

let solve_until env c meaning =
  let rec aux () = match c.constraints#get with
    | [] -> failwith "solve_until"
    | e :: q ->
        if e.edge_meaning <> meaning then (
          c.exists#set [];
          c.constraints#set q;
          fast_solve_one env c e;
          aux ()
        )
  in aux ()


(* Record containing the nodes of a an application constraint.
   The nodes for fun and arg are supposed to be the "real" root
   of the type, not the forall nodes *)
type app_nodes = {
  node_fun: node;
  node_arg: node;
  node_app: node;
}

let extract_app_fun_arg c = match c.constraints#get with
  | [] | [_] -> assert false
  | earg :: efun :: _ ->
      let extract_node e = match e with
        | { edge_type = Unif; edge_nodes = (n, _) } -> n
        | { edge_type = Inst; edge_nodes = (n, _) } -> child1 n
      in {
        node_fun = extract_node efun;
        node_arg = extract_node earg;
        node_app = binder (snd (efun.edge_nodes))
      }

(* Makes the meaning of a constraint more precise. Right now, we
flag applications of terms which are not function *)
let rewrite_meaning e m c =
  match e, m with
    | Unif.Clash ((_, _), (_, `Arrow)), CApp (loc, _)
        when (let nfun = (extract_app_fun_arg c).node_fun in
              !!nfun.constr <> Constr `Arrow) ->
        CNotFun loc
    | _ -> m
