open GraphicTypes
open Nodes
open GraphicSyntacticTranslations
open Constraints
open ConstraintsSolving
open ConstraintsMeaning
open Highlight
open Format


(* Exceptions resulting from the solving of constraints, after having
   been replayed. We give access to the constraint before the actions
   which caused the error in the first place *)
exception InferenceErrorReplayed of
  Unif.unif_error * constraint_meaning * constraints


let print_type (tr : graphic_translation) ppf n =
  let t = tr.node_type n in
  fprintf ppf "%a%s"
    (fun ppf -> SyntacticTypesPrinting.print_polytype_no_alpha ppf) t
    (match !!n.bound with
    | Root -> ""
    | Bound { permissions = PFlex } -> ""
    | Bound { permissions = PRigid } -> " (in a rigid context)"
    | Bound { permissions = PLocked } -> " (in a locked context)"
    )

let print_context (tr : graphic_translation) n ppf =
  let l = context_above_node tr n in
  if l <> [] then (
    fprintf ppf "@.@[The@ context@ is@ ";
    Utils.print_list
      (fun ppf (v, b, t) -> fprintf ppf "('%s %a %a)"
         (TypeConstructors.typevar_s v)
         SyntacticTypesPrinting.print_binding_label b
         (fun ppf -> SyntacticTypesPrinting.print_polytype_no_alpha ppf) t
      ) " " ppf l;
    fprintf ppf "@]"
  )


let reset_app_constraints c = match c.constraints#get with
  | [] | [_] -> assert false
  | earg :: efun :: _ ->
      let expand e = match e.edge_type with
        | Unif -> None
        | Inst ->
            let n = fst e.edge_nodes in
            reset_forall n;
            if Nodes.equal n (binder (child1 n)) then
              Some (child1 n)
            else
              None
      in
      expand earg :: expand efun :: []


type app_printing = {
  nodes: app_nodes;
  tr: graphic_translation;
}

(* Printing of types coming from an app node in error messages *)
let print_aux c =
  let roots = reset_app_constraints c in
  let nodes = extract_app_fun_arg c in
  let in_domain n =
    List.exists (function None -> false | Some n' ->
                   GraphicTypesAux.transitively_bound n' n)
      (Some nodes.node_app :: roots)
  in
  let tr = graph_to_syntactic_aux ~in_domain ~nf:true () in {
    nodes = nodes;
    tr = tr;
  }

(* Helper function to print respectively one and two types,
   with the context afterwards. Impossible to factor in a list,
   as it would require formats with a variable number of holes... *)
let print_one_type format print ppf n =
  let napp = print.nodes.node_app
  and tr = print.tr in
  fprintf ppf format (print_type tr) n;
  print_context tr napp ppf

let print_two_types format print ppf (n1, n2) =
  let napp = print.nodes.node_app
  and tr = print.tr in
  let print_type' = print_type tr in
  fprintf ppf format print_type' n1 print_type' n2;
  print_context tr napp ppf


(* Printing of the argument of an application *)
let print_type_arg f ppf c =
  let p = print_aux c in
  print_one_type f p ppf p.nodes.node_arg

(* Printing of the function on an application *)
let print_type_fun f ppf c =
  let p = print_aux c in
  print_one_type f p ppf p.nodes.node_fun

(* Printing of the two arguments of an application *)
let print_type_app f ppf c =
  let p = print_aux c in
  print_two_types f p ppf (p.nodes.node_fun, p.nodes.node_arg)

let print_type_two_args f ppf c =
  let p = print_aux c in
  assert (!!(p.nodes.node_fun).constr = Constr `Arrow);
  let nfun' = fst (child2 p.nodes.node_fun) in
  print_two_types f p ppf (nfun', p.nodes.node_arg)


let do_if f = function
  | None -> ()
  | Some c -> f c

let rec constraint_meaning m e ?c ppf = match m with
  | CAbs _ | CUple _ | CEqFirst _ | CSeqNonUnit _ | CIfThenWithElse _
  | CEndSeq _  | CExistsVar _ -> assert false

  | CRaise m -> constraint_meaning m e ?c ppf

  | CVar (_loc_use, loc_bound) ->
      (* This should never raise an error, but just in case *)
      print_loc ppf loc_bound;
      fprintf ppf "@[This@ variable@ is@ used@ with@ incompatible@ types.@]";

  | CInt loc ->
      print_loc ppf loc;
      fprintf ppf "@[This@ expression@ should@ have@ type@ int.";
      do_if (print_type_arg "@ Here,@ it@ has@ type@ %a.@]" ppf) c;

  | CNotFun loc ->
      print_loc ppf loc;
      fprintf ppf
        "@[This@ expression@ is@ not@ a@ function,@ it@ cannot@ be@ applied.";
      do_if (print_type_fun " (It@ has@ type@ %a)@]" ppf) c;

  | CIfCond loc ->
      print_loc ppf loc;
      fprintf ppf "@[This@ condition@ should@ have@ type@ bool.";
      do_if (print_type_arg "@ Here,@ it@ has@ type@ %a.@]" ppf) c;

  | CIfMissingElse (loc_expr, _loc_if) ->
      print_loc ppf loc_expr;
      fprintf ppf "@[This@ expression@ should@ have@ type@ unit:@ \
there@ is@ no@ 'else'@ clause@ in@ the@ englobing@ 'if'.";
      do_if (print_type_arg "@ Here@ it@ has@ type@ %a.@]" ppf) c;

  | CIfEq (_loc_then, _loc_else, loc_if) ->
      print_loc ppf loc_if;
      fprintf ppf
        "@[Both@ branches@ of@ this@ 'if'@ have@ incompatible@ types.@]@.";
      do_if (print_type_two_args "@[The@ 'then'@ part@ has@ type@ %a@ \
while@ the@ 'else'@ part@ has@ type@ %a.@]" ppf) c;

  | CSeqUnit loc ->
      print_loc ppf loc;
      fprintf ppf "@[This@ statement@ should@ have@ type@ unit.";
      do_if (print_type_arg "@ Here@ it@ has@ type@ %a.@]" ppf) c;

  | CEq (loc1, loc2) ->
      print_locs ppf loc1 loc2;
      fprintf ppf "@[These@ two@ expressions@ have@ incompatible@ types.@]@.";
      do_if (print_type_two_args "@[The@ first@ expression@ has@ type@ %a@ while@ the@ second@ has@ type@ %a.@]" ppf) c;

  | CAppCoercion (loc_t, loc_expr) ->
      print_locs ppf loc_expr loc_t;
      fprintf ppf
        "@[The@ expression@ at@ the@ first@ location@ cannot@ be@ coerced@ to@ the@ type@ at@ the@ second:@ %a."
        (Unif.print_unif_error_explanation "expression") e;
      do_if (print_type_arg "@ The@ expression@ has@ type@ %a.@]" ppf) c;

  | CApp (loc_fun, loc_arg) ->
      print_locs ppf loc_fun loc_arg;
      fprintf ppf
        "@[Cannot@ apply@ the@ first@ expression@ to@ the@ second:@ %a.@]@."
        (Unif.print_unif_error_explanation "argument") e;
      do_if (print_type_app "@[The@ first@ expression@ has@ type@ %a@ while@ the@ second@ has@ type@ %a.@]" ppf) c;

  | CFix loc ->
      print_loc ppf loc;
      fprintf ppf "@[Cannot@ typecheck@ this@ recursive@ expression:@ %a.@]@."
        (Unif.print_unif_error_explanation "body") e;
      fprintf ppf "@[Try@ adding@ a@ type@ annotation@ around@ the@ function@ name@ to@ get@ better@ error@ messages.@]@.";
      do_if (print_type_arg "@[The@ inferred@ type@ for@ the@ functional@ of@ the@ recursive@ function@ is@ %a,@ which@ should@ be@ an@ instance@ of@ 'a. ('a -> 'a) -> 'a.@]" ppf) c;
;;
