(*pp camlp4o ./hashtbl_ext.cmo *)

open TypeConstructors
open SyntacticTypes
open SyntacticTypes.KnownconstrTypes
open GraphicTypes
open GraphicTypes.Nodes
open GraphicTypesAux


(* This is called ftv in Didier Le Botlan's phd *)
let rec used_var var : _ polytype_ -> bool = function
  | `Var var' -> var = var'
  | `Bottom -> false
  | `Constr (_, ls) -> List.exists (used_var var) (ls :> _ polytype_ list)
  | `Bound (_, (var', bound), s) ->
      if var = var' then
        used_var var s
      else
        used_var var s || (used_var var bound && used_var var' s)


let () = #{Hashtbl} (* Used on typevar *)

(* Translation of a monotype into a graph. We use an environment
   which must bind all the variables to a node. The monomorphic
   nodes are bound using the provided bound *)
let rec monotype_to_graph ~bound ~env : monotype -> node = function
  | `Var v -> env{v}
  | `Constr (c, l) ->
      let l' = List.map (monotype_to_graph ~bound ~env) l in
      let n = new_node (Constr c) l' in
      set_binder n bound;
      n
(* Translation of a monotype require a permissions system and the
   current permissions of the node *)
and polytype_to_graph ~perms ~env ~bheight : polytype -> node = function
  | #monotype as t ->
      (* Temporary root for the monotype *)
      let r = new_node Bot [] in
      let flag = match perms with
        | PFlex -> Flex
        | PRigid | PLocked -> Rigid
      in
      (* Bound for the temporary root. Note that it is cyclic,
         hence it is essential to overwrite it later. This is done
         for the root (by setting the bound to Root in syntactic_to_graph),
         and for the bounds of quantified variables *)
      let bound = { label = flag; binder = r; binding_height = bheight;
                    permissions = permissions_syst perms flag } in
      let n = monotype_to_graph ~bound ~env t in
      fuse ~keep:n ~lose:r ();
      n
  | `Bottom -> new_node Bot []
  | `Bound (blabel, (var, t), t') ->
      if used_var var t' then (* Rule Eq-Free iterated *)
        let bperms = permissions_syst perms blabel in
        let nbound = polytype_to_graph ~env ~perms:bperms t
          ~bheight:(bheight+1) in
        env{var} <+ nbound;
        let nt' = polytype_to_graph ~perms ~env ~bheight t' in
        if Nodes.equal nt' nbound = false then (* Rule Eq-Var *)
          set_binder nbound ({label = blabel; binder =  nt';
                              permissions = bperms; binding_height = bheight});
        env{var} ->;
        nt'
      else
        polytype_to_graph ~perms ~env ~bheight t'

(* Raised in case of un unbound variable *)
exception UnboundTypeVar

(* Translation of a syntactic type into a graph. The type seen
   as a polytype is translated (with the root permission being
   Flex).  *)
let syntactic_to_graph s =
  try
    let env = {} in
    let n = polytype_to_graph ~env ~perms:PFlex ~bheight:1 s in
    !!n.bound <- Root;
    n
  with Not_found -> raise UnboundTypeVar



exception ContainsConstraints

let () = #{HashNodes}

type context = (typevar * binding_label * polytype) list

let rebind_context t (c : context) =
  List.fold_left (fun t (v, l, b) ->
                    match l with
                      | Flex -> SyntacticTypes.flex v ~b t
                      | Rigid -> SyntacticTypes.rigid v ~b t)
    t c

type graphic_translation = {
  node_type : node -> polytype;
  node_context: node -> context
}

(* Translation from graphs to syntactic types.
The translation can deal with nodes that are not in the scope
of a node. It returns an object with two methods, one
for the type of a node, one for the context of a node
(which is defined as the set of nodes not in the scope
bound on the argument node)
 *)
let graph_to_syntactic_aux ?(nf=true) ?(in_domain = fun _ -> true) () : graphic_translation =
  (* (constant) name for a node *)
  let name =
    let c = Utils.counter 0
    and names = {} in
    fun n ->
      try names{n}
      with Not_found ->
        let v = var_n c#get in
        names{n} <+ v;
        v
  in

  (* Nodes bound on a node, with the name of the bound, the flag,
     and the bound itself *)
  let bound_on = {} in

  (* Cached translation as a monotype *)
  let tr = {} in

  let context n : context = try bound_on{n} with Not_found -> []
  and translate n =
    let is_mono = if nf then mono n else not_mono in

    let bind n t =
      let v = name n in
      (match !!n.bound with
         | Root -> ()
         | Bound { binder = n'; label = b } ->
             let c = (v, b, t) in
             try
               let l = bound_on{n'} in
               bound_on{n'} <- c :: l
             with Not_found -> bound_on{n'} <- [c]
      );
      tr{n} <- `Var v;
      `Var v
    in

    let rec as_var_or_mono n =
      try tr{n}
      with Not_found ->
        if is_mono n then
          as_mono n
        else
          let t = as_poly n in
          bind n t
    and as_poly n = match !!n.constr with
      | Forall _ -> raise ContainsConstraints
      | Bot -> `Bottom
      | Type t -> t
      | Constr c  ->
          let l' = List.map as_var_or_mono !!n.children in
          let mono = `Constr (c, l') in
          try
            let bound_on = bound_on{n} in
            rebind_context mono bound_on
          with Not_found -> mono
    and as_mono n =
      try tr{n}
      with Not_found ->
        match !!n.constr with
          | Forall _ -> raise ContainsConstraints
          | Bot -> bind n `Bottom
          | Type (#monotype as t) when really_monomorphic_monotype t -> t
          | Type t -> bind n t
          | Constr c ->
              let l' = List.map as_var_or_mono !!n.children in
              let t = `Constr (c, l') in
              tr{n} <- t;
              t
    in
    if in_domain n then
      if !!n.constr = Bot then `Bottom else as_poly n
    else
      (as_var_or_mono n :> polytype)
  in {
    node_type = translate;
    node_context = context;
  }


let rec context_above_node (tr : graphic_translation) n : context =
  let rec aux n l =
    let l' = tr.node_context n @ l in
    match !!n.bound with
      | Root -> l'
      | Bound { binder = n' } -> aux n' l'
  in
  aux n []


let graph_to_syntactic ?(nf=true) r =
  (graph_to_syntactic_aux ~nf ()).node_type r
