(*pp camlp4o ./hashtbl_ext.cmo *)

open TypeConstructors
open Utils
open Format


type forall_use = { mutable forall : int }

type gconstr =
  | Constr of typeconstr
  | Type of SyntacticTypes.KnownconstrTypes.polytype
  | Bot
  | Forall of forall_use

let is_forall = function
  | Forall _ -> true
  | _ -> false

let print_gconstr ?(output=(!Options.output)) () = function
  | Bot -> SyntacticTypesPrinting.string_of_bottom ~output ()
  | Constr c -> string_of_typeconstr ~output c
  | Type t ->
      format_into_string SyntacticTypesPrinting.print_polytype t
  | Forall _ -> "G"
      (*
      match output with
        | Options.LaTeX -> "\\forall"
        | Options.Unicode -> "&#8704;"
        | Options.Ascii -> "forall"
      *)

let garity = function
  | Bot -> 0
  | Forall _ -> 1
  | Constr c -> arity c
  | Type _ -> 0

let no_mono = function
  | Bot -> true
  | Forall _ -> true
  | Constr c -> typeconstr_is_poly c
  | Type t -> SyntacticTypes.polymorphic_type t


type binding_label = SyntacticTypes.syntactic_binding_label = Flex | Rigid

module NId : sig
  type id
  type t = id
  val fresh : unit -> id
  val compare : id -> id -> int
  val to_int : id -> int
  val hash : id -> int
  val equal : id -> id -> bool

  val counter_state : unit -> id
  val reset_counter : id -> unit
end =
struct
  type id = int
  type t = id
  let counter = counter 0
  let fresh () = counter#get
  let compare id1 id2 = id1 - id2
  let to_int (i:id) = i
  let hash (i:id) = i
  let equal (i1:id) (i2:id) = i1 = i2

  (* Not for the casual user ... *)
  let counter_state () = counter#peek
  let reset_counter v = counter#set_to v
end


(* Computations of the permissions of a node *)
type permissions = PFlex | PRigid | PLocked


let permissions_syst p l =
  match l with
    | Rigid -> PRigid
    | Flex -> if p = PFlex then PFlex else PLocked

let print_permissions = function
  | PFlex -> "green"
  | PRigid -> "orange"
  | PLocked -> "red"


(* Types of nodes and associated operations *)
module Nodes = struct
  type node = {
    node_info: node_info UnionFind.point;
    old_struct: old_struct UnionFind.point;
  }
  and node_info = {
    id: NId.id;
    constr: gconstr;
    children: node list;
    mutable bound: node bound;
    canonical_node: node;
    unif_info: unif_info;
  }
  and 'a bound = Root | Bound of 'a bound_info
  and 'a bound_info = {
    label: binding_label;
    binder: 'a;
    permissions: permissions;
    binding_height: int;
  }
  and unif_info = {
    mutable unif_tag: [`Unif] Tag.tag;
    mutable unified_nodes: (NId.id * node bound) Tree.tree;
    mutable bot_non_bot: bool;
    mutable merge_took_place: bool;
  }
  and old_struct = {
    old_id: NId.id;
    old_children: node list;
  }

  let (!!) n = UnionFind.find n.node_info
  let (!!.) n = UnionFind.find n.old_struct

  type t = node

  let equal x y =
    NId.compare !!x.id !!y.id = 0

  let compare x y =
    NId.compare !!x.id !!y.id

  let hash x =
    NId.hash !!x.id

end


open Nodes


let new_node c lcn =
  (* We make sure to obtain a new forall count *)
  let c = match c with
    | Forall { forall = d } -> Forall ({ forall = d })
    | c -> c
  in
  let ui = { (* Those fields will be rewritten at the beginning of each
                unification, the default values have no real meaning.
              On the other hand, the field unif_info itself
              is unmutable, so we must create a new record for
              each node. *)
    unif_tag = Tag.dummy_tag;
    unified_nodes = Tree.Empty;
    bot_non_bot = false;
    merge_took_place = false;
  } in
  let id = NId.fresh () in
  let nos = {
    old_id = id;
    old_children = lcn;
  } in
  let ni' = Obj.magic () in
  let rec ni = {
    id = id;
    constr = c;
    children = lcn;
    bound = Root;
    unif_info = ui;
    canonical_node = n;
  }
  and n = {
    node_info = ni';
    old_struct = UnionFind.fresh nos;
  }
  in Obj.set_field (Obj.repr n) 0 (Obj.repr (UnionFind.fresh ni));
  n


let nid n = NId.to_int !!n.id


let set_binder n bound =
  !!n.bound <- Bound bound


let is_root r = !!r.bound = Root

(*
let is_bound n = match !!n.bound with
  | Root | Unbound -> false
  | Bound _ -> true
*)

let _bound n = match !!n.bound with
  | Root -> assert false
  | Bound b -> b

let binder n = (_bound n).binder
let blabel n = (_bound n).label
let permissions n =
  if is_root n then PFlex else (_bound n).permissions

let binder_option n = match !!n.bound with
  | Bound {binder = b} -> Some b
  | _ -> None

let binding_height n = match !!n.bound with
  | Root -> 0
  | Bound {binding_height = h} -> h

let children n = !!n.children

let child1 n = match !!n.children with
  | [n] -> n
  | _ -> assert false

let child2 n = match !!n.children with
  | [n1; n2] -> n1, n2
  | _ -> assert false


let bind_to ?binder ~n l =
  !!n.bound <- match binder with
    | None -> Root
    | Some n' -> Bound ({
        label = l;
        binder = n';
        permissions = permissions_syst (permissions n') l;
        binding_height = binding_height n'+1;
      })

let recompute_node_infos n =
  match !!n.bound with
    | Root -> ()
    | Bound { binder = n'; label = l } ->
        bind_to ~binder:n' ~n l


(* Used to fuse a dummy node with a real one. No checks are
   performed *)
let fuse ~keep ~lose ?(bound=keep) () =
  let bound = !!bound.bound in
  UnionFind.union ~keep:keep.node_info ~lose:lose.node_info ();
  !!keep.bound <- bound;
  UnionFind.union ~keep:keep.old_struct ~lose:lose.old_struct ();


(* Standard structures on nodes *)
module SetNodes = Set.Make(Nodes)
module HashNodes = Hashtbl.Make(Nodes)

