(*pp camlp4o ./hashtbl_ext.cmo *)

open GraphicTypes
open Nodes

let () = #{HashNodes}

(* General DFS functions on nodes *)
type dfs = {
  has_cycle: ?return_cycle:bool -> node -> bool;
  generic_dfs: ?pre:(node -> unit) -> ?post:(node -> unit) -> node -> unit;
  topological_sort_list : node -> node list;
  topological_sort_index : node -> int HashNodes.t;
}


exception Cycle of node list

let aux_dfs succs =

(* invariant: not in [h] means not visited at all; [h v = true] means
   already visited in the current component; [h v = false] means
   already visited in another tree *)
  let rec has_cycle ?(return_cycle=false) n =
    let h = {} in
    let rec visit preds n =
      h{n} <+ true;
      List.iter
        (fun n' ->
           try if h{n'} then raise (Cycle (n :: preds))
           with Not_found -> visit (n :: preds) n'
        ) (succs n);
      h{n} <+ false
    in
    try List.iter (fun v -> if h{?v} = false then visit [v] v) (succs n); false
    with Cycle _ as e -> if return_cycle then raise e else true

  and generic_dfs ?(pre=fun _ -> ()) ?(post=fun _ -> ()) n =
    let h = {} in
    let rec visit n =
      if h{?n} = false then begin
        h{n} <+ ();
        pre n;
        List.iter visit (succs n);
        post n
      end
    in
    visit n;

  and topological_sort_list n =
    let l = ref [] in
    let post n =
      l := n :: !l
    in
    assert (has_cycle n = false);
    generic_dfs ~post n;
    !l;

  and topological_sort_index n =
    let l = topological_sort_list n in
    let r = {} in
    ignore (List.fold_left (fun i n -> r{n} <+ i; i+1) 0 l);
    r;

  in {
    has_cycle = has_cycle;
    generic_dfs = generic_dfs;
    topological_sort_list = topological_sort_list;
    topological_sort_index = topological_sort_index;
  }

let dfs_struct = aux_dfs children



(* Invert the edges of a graph. *)
type inverse_graph = {
  mutable struct_reach: SetNodes.t;
  mutable bound_at: node list;
}

let aux_compute_inverse_graph
    (dfs : ?pre:(node -> unit) -> ?post:(node -> unit) -> node -> unit)
    ?(compute_struct=true) ?(compute_binding=true) r =
  let h = {} in
  let res n =
    try h{n}
    with Not_found ->
      let v = {
        struct_reach = SetNodes.empty;
        bound_at = [];
      } in
      h{n} <- v;
      v
  in
  let aux n =
    ignore (res n);
    if compute_struct then
      List.iter (fun n' ->
                   let s = res n' in
                   s.struct_reach <- SetNodes.add n s.struct_reach)
        (children n);
    if compute_binding then
      match !!n.bound  with
        | Root -> ()
        | Bound { binder = b } ->
            let s = res b in
            s.bound_at <- n :: s.bound_at
  in
  dfs ~pre:aux r;
  h

let compute_inverse_graph = aux_compute_inverse_graph dfs_struct.generic_dfs


let tarjan succs n =
  let dfs = aux_dfs succs in
  let inverse = aux_compute_inverse_graph dfs.generic_dfs n in
  let preds n = match binder_option n with
    | None -> inverse{n}.struct_reach
    | Some n' -> SetNodes.add n' inverse{n}.struct_reach
  and h = {}
  and c = Utils.counter(-1) in
  dfs.generic_dfs ~pre:(fun n -> h{n} <- c#get) n;
  let number = c#get in
  let module Aux = struct
    type node = GraphicTypes.Nodes.node

    let start = n

    let n = number
    let index n = h{n}

    let successors f n = List.iter f (succs n)
    let predecessors f n = SetNodes.iter f (preds n)
  end in
  let module TL = Lengauer.Run(Aux) in TL.dominator


(* Annotates the nodes under n according to whether or not
   they are monomorphic *)
let mono ?(dfs=dfs_struct) n =
  let used = {}
  and r = {} in
  let aux n =
    let ni = !!n in
    if used{?n} || no_mono ni.constr then
      match ni.bound with
        | Root -> ()
        | Bound { binder = n' } -> used{n'} <+ ()
    else
      r{n} <+ ()
  in
  dfs.generic_dfs ~post:aux n;
  (fun n -> r{?n})

let not_mono (_ : node) = false


(* Deep copy of a type. All nodes are copied into new ones using
   new_node *)
let copy r =
  assert (is_root r);
  let copy = {} in
  let repr n = copy{n} in
  let copy_node n =
    copy{n} <+ new_node !!n.constr (List.map repr !!n.children)
  and update_copy n =
    let n' = copy{n} in
    match !!n.bound with
      | Root ->  !!n'.bound <- Root
      | Bound bi ->
          !!n'.bound <- Bound ({bi with binder = repr bi.binder})
  in
  dfs_struct.generic_dfs ~post:copy_node r;
  dfs_struct.generic_dfs ~pre:update_copy r;
  copy{r}


(* Used to sort the binders of n by binding height *)
module SetNodesHeight = Set.Make(
  struct
    type t = node

    let compare n1 n2 =
      let h1 = binding_height n1 and h2 = binding_height n2 in
      if h1 = h2 then
        Nodes.compare n1 n2
      else
        (* The node bound higher must come last, as it is
           the other one that must be raised *)
        h2 - h1
  end)

(* Find the lca of a set of nodes by recursively raising the lowest
   bound node *)
let rec lca_binders a =
  let lowest_bound_node = SetNodesHeight.min_elt a in
  let a' = SetNodesHeight.remove lowest_bound_node a in
  if SetNodesHeight.is_empty a' then
    lowest_bound_node
  else
    (* lowest_bound_node must be bound, as we visit the nodes
       in an order given by a topological sort, and only the
       root can be unbound.  But if lowest_bound_node is the
       root, it must be alone in ancestors *)
    let n = binder lowest_bound_node in
    lca_binders (SetNodesHeight.add n a')

let rec transitively_bound b n =
  Nodes.equal b n ||
    match !!n.bound with
      | Bound { binder = n' } -> transitively_bound b n'
      | Root -> false
