open TypeConstructors


type syntactic_binding_label = Flex | Rigid

(** MLF monotypes *)
type mono_var = [`Var of typevar]
type ('a, 'b) sconstr = [`Constr of 'b * 'a list]
type 'a monotype_ = [ mono_var | ('a monotype_, 'a) sconstr]

type ('a, 'b) bound =
    [ `Bound of syntactic_binding_label * (typevar * 'b) * 'a ]

(** The unconstrained, standard MLF types *)
type 'a polytype_ = [
  | 'a monotype_
  | `Bottom
  | ('a polytype_, 'a polytype_) bound
]


(* Polytypes in which the usual monotypes can also be polytypes *)
type 'a expanded_polytype_ = [
  | mono_var
  | ('a expanded_polytype_, 'a) sconstr
  | `Bottom
  | ('a expanded_polytype_, 'a expanded_polytype_) bound
]

let rec ftv' : _ expanded_polytype_ -> _ = function
  | `Var v -> SetTypeVars.singleton v
  | `Bottom -> SetTypeVars.empty
  | `Constr (_, ls) ->
      List.fold_left
        (fun s t -> SetTypeVars.union s (ftv' t))
        SetTypeVars.empty ls
  | `Bound (_, (v, bound), ret) ->
      SetTypeVars.union (SetTypeVars.remove v (ftv' ret)) (ftv' bound)

(* All the types variables (included the bound ones) of a type *)
let rec atv : _ expanded_polytype_ -> _ = function
  | `Var v -> SetTypeVars.singleton v
  | `Bottom -> SetTypeVars.empty
  | `Constr (_, ls) ->
      List.fold_left
        (fun s t -> SetTypeVars.union s (atv t))
        SetTypeVars.empty ls
  | `Bound (_, (v, bound), ret) ->
      SetTypeVars.union (SetTypeVars.union (atv ret) (atv bound))
        (SetTypeVars.singleton v)

let ftv t = ftv' (t : 'a polytype_ :> 'a expanded_polytype_)


let flex n ?(b=`Bottom) a : [> _ polytype_] =
  `Bound (Flex, (n, b), a)
let rigid n ?(b=`Bottom) a : [> _ polytype_] =
  `Bound (Rigid, (n, b), a)


let arrow t1 t2  : [> _ monotype_] = `Constr (`Arrow, [t1 ; t2])
let var v : [> mono_var] = `Var (s_typevar v)

let var_n n =
  assert (n > 0);
  s_typevar (
    if n <= 26 then
      String.make 1 (char_of_int (97-1+n))
    else
      "a" ^ (string_of_int (n-26))
  )


(* Sugar informations *)
type sugar = syntactic_binding_label option

let compatible_sugar sugar lbl =
  sugar = Some lbl

let sugar_arrow = [Some Rigid; Some Flex]
let sugar_uple b = let rec v = b :: v in v
let rec sugar_unknown = None :: sugar_unknown


type sugar_env = {
  sugar_arrow: sugar list;
  sugar_uple: sugar list;
  sugar_inductive: (user_type_desc * sugar list) list;
}

let default_sugar_env = {
  sugar_arrow = sugar_arrow;
  sugar_uple = sugar_uple (Some Flex);
  sugar_inductive = [];
}

let no_sugar = {
  sugar_arrow = [None; None];
  sugar_uple = sugar_unknown;
  sugar_inductive = [];
}

let sugar_constr env (c : typeconstr) = match c with
  | `Arrow -> env.sugar_arrow
  | `Uple _ -> env.sugar_uple
  | #init_ct_typeconstr -> []
  | `User d ->
      try List.assoc d env.sugar_inductive
      with Not_found -> sugar_unknown



(* All bounds collected in a prefix *)
type 'a prefix_bound = syntactic_binding_label * (typevar * 'a polytype_)
type 'a prefix = 'a prefix_bound list


(* Variables not already present in a type *)
let fresh_var avoid =
  let c = Utils.counter 0 in
  let rec new_var () =
    let s = var_n c#get in
    if SetTypeVars.mem s avoid then
      new_var ()
    else
      s
  in new_var


module KnownconstrTypes = struct
  type monotype = typeconstr monotype_
  type polytype = typeconstr polytype_
  type expanded_polytype = typeconstr expanded_polytype_
end

module UnknownconstrTypes = struct
  type monotype = unresolved_typeconstr monotype_
  type polytype = unresolved_typeconstr polytype_
  type expanded_polytype = unresolved_typeconstr expanded_polytype_
end


open KnownconstrTypes


let rec really_monomorphic_monotype : monotype -> bool = function
  | `Var _ -> false
  | `Constr ((#init_ct_typeconstr | `Arrow | `Uple _), l) ->
      (* Builtin types are inductive types, not aliases *)
      List.for_all really_monomorphic_monotype l
  | `Constr (`User c, l) ->
      c.typeconstr_is_poly = false &&
      List.for_all really_monomorphic_monotype l

let rec polymorphic_type : polytype -> bool = function
  | #monotype as t -> not (really_monomorphic_monotype t)
  | #polytype -> true


let unexpand sugar_env (t : expanded_polytype) =
  let new_var = fresh_var (atv t) in

  let rec unexpand : expanded_polytype -> polytype = function
    | `Var _ | `Bottom as c -> c
    | `Constr (c, l) -> unexpand_constr c l
    | `Bound (l, (v, t), t') ->
        `Bound (l, (v, unexpand t), unexpand t')
  and unexpand_constr c l =
    let sugar = sugar_constr sugar_env c in
    let b, m = unexpand_monotype_list sugar l in
    rebind (`Constr (c, m)) b
  and unexpand_monotype_list sugar l : _ prefix * _ = match l, sugar with
    | [], _ -> [], []
    | p :: q, v :: qv ->
        let b, m = unexpand_monotype v p in
        let lb, lm = unexpand_monotype_list qv q in
        b @ lb, m :: lm
    | _ :: _, [] -> assert false
  and unexpand_monotype sugar : expanded_polytype -> _ prefix * monotype = function
    | `Var _ as v -> [], v
    | `Constr (c, l) ->
        (match unexpand_constr c l, typeconstr_is_poly c with
           | #monotype_ as m, false -> [], m
           | #polytype_ as p, _ -> poly_to_mono sugar p
        )
    | `Bottom  -> poly_to_mono sugar `Bottom
    | `Bound _ as b ->
        let b' = unexpand b in
        poly_to_mono sugar b'
  and poly_to_mono sugar poly =
    let v = new_var () in
    let lbl = if sugar = Some Flex then Flex else Rigid in
    [lbl, (v, poly)], `Var v
  and rebind t = function
    | [] -> t
    | (lbl, (b, t')) :: q ->
        let f = if lbl = Flex then flex else rigid in
        rebind (f b ~b:t' t) q
  in
  unexpand t
