open TypeConstructors
open SyntacticTypes
open Format


let string_of_bottom ?(output=(!Options.output)) () =
  match output with
    | Options.LaTeX -> "\\bot"
    | Options.Unicode -> "&#8869;"
    | Options.Ascii -> "_|_"


let print_binding_label ppf = function
  | Flex -> fprintf ppf ">"
  | Rigid -> fprintf ppf "="



let level_var = 0
let level_constr = 2
let level_uple = 4
let level_arrow = 6
let level_poly = 8


let level : _ expanded_polytype_ -> int = function
  | `Var _ | `Bottom -> level_var
  | `Constr (`Arrow, _) -> level_arrow
  | `Constr (`Uple _, _) -> level_uple
  | `Constr (_, _) -> level_constr
  | `Bound _ -> level_poly

let rec print_type ppf : _ expanded_polytype_ -> _ = function
  | `Var v ->
      let v = typevar_s v in
      assert (String.length v > 0);
      if v.[0] = '#' then (* Nameless variable, created at parsing type *)
        fprintf ppf "_"
      else
        fprintf ppf "'%s" v

  | `Constr (`Uple n, l) ->
      assert (n >= 2);
      Utils.print_list (print_type' level_uple) " * " ppf l

  | `Constr (`Arrow, [t1;t2]) ->
      fprintf ppf "%a %s %a"
        (print_type' level_arrow) t1
        (string_of_typeconstr ~output:Options.Ascii `Arrow)
        (print_type' (level_arrow + 1)) t2

  | `Constr (`Arrow, _) -> assert false

  | `Constr (c, []) ->
      fprintf ppf "%s" (string_of_typeconstr ~output:Options.Ascii c)

  | `Constr (c, [t]) ->
      fprintf ppf "%a %s"
        (print_type'
           (level_constr + if !Options.minimize_parentheses then 1 else 0)) t
        (string_of_typeconstr ~output:Options.Ascii c)

  | `Constr (c, l) ->
      fprintf ppf "(%a) %s"
        (Utils.print_list (print_type' level_poly) ", ") l
        (string_of_typeconstr ~output:Options.Ascii c)

  | `Bottom -> fprintf ppf "%s" (string_of_bottom ())

  | `Bound (Flex, (v, `Bottom), t) ->
      fprintf ppf "'%s. %a" (typevar_s v) print_type t

  | `Bound (Rigid, (v, `Bottom), t) ->
      fprintf ppf "('%s = %s). %a"
        (typevar_s v) (string_of_bottom ()) print_type t

  | `Bound (l, (v, t), t') ->
      fprintf ppf "('%s %a %a). %a"
        (typevar_s v) print_binding_label l print_type t print_type t'
and print_type' lvl ppf t =
  if level t >= lvl then
    fprintf ppf "(%a)" print_type t
  else
    fprintf ppf "%a" print_type t



(* In the following, we suppose that types are normalized : no bound
   variable used twice, no usage of Eq-Var possible, no dead bounds. *)

let sort_prefix_aux =
  let aux (l1, (_, b1)) (l2, (_, b2)) =
    match l1, b1, l2, b2 with
      | Flex, `Bottom, _, _ -> -1
      | Rigid, `Bottom, Flex, `Bottom -> 1
      | Rigid, `Bottom, _, _ -> 1
      | _, _, _, _ -> -1
  in (aux : 'a prefix_bound -> 'a prefix_bound -> int)


(* Split a polytype into a prefix and the type at the end *)
let rec split : _ polytype_ -> _ prefix * _ = function
  | `Bottom -> [], `Bottom
  | #monotype_ as t ->
      [], t
  | `Bound (l, b, t) ->
      let bounds, t = split t in
      (l, b) :: bounds, t


(* Find a variable in a prefix. Return the bound, the remainder of
   the prefix (without the bound) and the prefix after *)
let rec split_q v : _ prefix -> _ = function
  | [] -> raise Not_found
  | (l, (v', t)) as b :: q ->
      if v = v' then
        (l, t), q, q
      else
        let b', q', qa = split_q v q in
        b', b :: q', qa




(* Does a type variable occur in a polytype. We check for the case
   of the variable being (re)bound inside the type *)
let rec occur var : _ polytype_ -> bool = function
  | `Var var' -> var = var'
  | `Bottom -> false
  | `Constr (_, ls) -> List.exists (occur var) (ls :> _ polytype_ list)
  | `Bound (_, (var', bound), s) ->
      if var = var' then
        occur var bound
      else
        occur var s || occur var bound

(* Nb of occurrences of a variable into a monotype *)
let rec nb_occur v : _ monotype_ -> int = function
  | `Var v' -> if v = v' then 1 else 0
  | `Constr (_, l) -> List.fold_left (fun c t -> c + nb_occur v t) 0 l


(* Inline the variables into the arguments of a constructor when
   the sugar information are respected. The variable is inlined
   if it occurs exactly once in the prefix and the arguments of the
   constructor. Note that we suppose that the variable does not appear
   more than once in the prefix. The sugar argument can be longer
   than the constructor arguments (we use this as a small optimization
   using infinite recursive lists). Bottom nodes are not inlined, in
   order to avoid difficult to read types such as _|_ -> _|_
*)
let rec inline_list q (l : 'a monotype_ list) sugar =
  let (>::) e (l, q) = (e :: l, q) in
  let rec aux q = function
    | [], _ -> [], q
    | (t :: tl), (var :: qvar) ->
        let default = ref ((t :> _ polytype_) , q) in
        (match t with
           | `Var v ->
               (try
                  let (lbl, t), q', qa = split_q v q in
                  if t <> `Bottom &&
                    compatible_sugar var lbl &&
                    List.for_all (fun (_, (_, t)) -> occur v t = false) qa &&
                    List.fold_left (fun c t -> c + nb_occur v t) 0 l = 1
                  then
                    default := (t, q')
                with Not_found -> ())
           | `Constr _ -> ()
        );
        let t, q = !default in
        t >:: aux q (tl, qvar)
    | _, [] -> assert false
  in aux q (l, sugar)


(* Expansion of polytypes inside the monotypes part. We first
   collect all the relevant bounds and expand them *)
let rec inline sugar_env : _ polytype_ -> _ expanded_polytype_ = function
  | `Bottom | `Var _ | `Constr (_, _) as t ->
      (t : _ polytype_ :> _ expanded_polytype_)
  | `Bound _ as t ->
      let bounds, t = split t in
      let bounds' = List.sort sort_prefix_aux bounds in
      match t with
        | `Bottom | `Var _ ->
            (* Cannot happen if types are correctly normalized *)
            assert false
        | `Constr (c, l) ->
            inline_in_constr sugar_env c bounds' l (sugar_constr sugar_env c)

(* Auxiliary function for inline in constructors *)
and inline_in_constr sugar_env (c : 'a) (q : 'a prefix) l sugar =
  let l', q' = inline_list q l sugar in
  let t = `Constr (c, List.map (inline sugar_env) l') in
  rebind sugar_env t q'

(* Rebinding of the variables of the prefix that have not been inlined *)
and rebind sugar_env (t : _ expanded_polytype_) q =
  let rec aux t = function
    | [] -> t
    | (l, (v, tb)) :: q ->
        let t' = `Bound (l, (v, inline sugar_env tb), t) in
        aux t' q
  in aux t (List.rev q)


let () = #{Hashtbl}

let alpha_convert t =
  let new_var = fresh_var (ftv' t) in
  let h = {} in
  let rec aux : _ expanded_polytype_ -> _ expanded_polytype_ = function
    | `Var v -> `Var (try h{v} with Not_found -> v)
    | `Constr (c, l) -> `Constr (c, List.map aux l)
    | `Bottom -> `Bottom
    | `Bound (l, (v, tb), t) ->
        let tb' = aux tb in
        let v' = new_var () in
        h{v} <+ v';
        let t' = aux t in
        h{v} ->;
        `Bound (l, (v', tb'), t')
  in
  aux t


let print_polytype_aux sugar_env alpha ppf t =
  let t = if true (* !Options.inline_types *) then
    inline sugar_env t (* Printing of types after expansion *)
  else
    (t: _ polytype_ :> _ expanded_polytype_) (* Non-inlined printing of types *)
  in
  let t = if alpha then alpha_convert t else t in
  print_type ppf t

let print_polytype ?(sugar_env=no_sugar) ppf =
  print_polytype_aux sugar_env true ppf
let print_polytype_no_alpha ?(sugar_env=default_sugar_env) ppf =
  print_polytype_aux sugar_env false ppf
