open TypingEnv
open Constraints
open ConstraintsGeneration
open ConstraintsSolving
open AST
open Locations
open Format


type input = {
  lexbuf: Lexing.lexbuf;
  source: string;
}

let desugar e =
(*  eprintf "@[%a@]@." print_expr (e : source_expr loc :> expr loc); *)
  let e = desugar e in
(*  eprintf "@[%a@]@." print_expr (e : typing_expr loc :> expr loc); *)
  e


let eat_semisemi_spaces source pos =
  let l = String.length source in
  let rec aux pos =
    if pos >= l then
      assert false
    else
      let c = source.[pos] in
      if c = ' ' || c = '\n' || c = '\r' || c = '\t' then
        aux (pos+1)
      else
        if pos < l-1 then
          if String.sub source pos 2 = ";;" then
            aux (pos+2)
          else
            pos
        else
          pos
  in aux pos

let format_source =
  ref (fun (x : (string -> unit, Format.formatter, unit) format) -> x)


let reprint_source ?ppf pos_start pos_end input =
  match ppf with
    | None -> ()
    | Some ppf ->
        if input.source <> "" then
          let pos_start = eat_semisemi_spaces input.source pos_start in
          let s = String.sub input.source pos_start (pos_end - pos_start)
          in fprintf ppf ("@.@[<hov 0>" ^^ (!format_source "# %s") ^^ "@]@.@.") s

let get_env_options () =
  !Highlight.input_lexbuf,
  !Highlight.input_name,
  !Dtypes.dump_ppf
;;

let restore_env_options (s1, s2, s5) =
  Highlight.input_lexbuf := s1;
  Highlight.input_name := s2;
  Dtypes.dump_ppf := s5;
;;


let solve_expr ?(nf=true) ?(check=no_check) ~env expr =
  let c, roots = translate env expr in
  try
    solve ~env ~check c;
    Dtypes.print_expr_roots env.sugar roots#get;
    forall_printing ~nf c, c
  with InferenceError (e, m) ->
    let c, _ = translate env expr in
    solve_until env c m;
    let m' = rewrite_meaning e m c in
    raise (ConstraintsSolvingErrors.InferenceErrorReplayed (e, m', c))

let fprintf' ppf f =
  match ppf with
    | None -> ifprintf std_formatter f
    | Some ppf -> fprintf ppf f

let std_ppf = function
  | None -> std_formatter
  | Some ppf -> ppf


let dump_svg = ref false

let solved_constraint_to_svg c =
  assert (Constraints.solved c);
  reset_forall c.root;
  GraphicTypesToDot.to_svg
    (fun ch -> GraphicTypesToDot.print_constraint ch ~skip_root:true c)

let constraint_to_svg c =
  let c = {c with exists = existentials_from_edges c} in
  GraphicTypesToDot.to_svg
    (fun ch -> GraphicTypesToDot.print_constraint ~show_mono:true ch c)


let rec type_phrase ?ppf env = function
  | Expr (e, success) ->
      let t, c = solve_expr ~env (desugar e) in
      fprintf' ppf "@[- :@ %a@]@."
        (SyntacticTypesPrinting.print_polytype ~sugar_env:env.sugar) t;
      if success = false then fprintf' ppf
        "@[Warning:@ the typing@ of this expression@ should@ have@ failed@]@.";
      if !dump_svg then fprintf' ppf "@[%s@]@." (solved_constraint_to_svg c);
      env

  | Val (r, v, e) ->
      let e = desugar (val_to_let r v e) in
      let t, c = solve_expr ~env e in
      let name = (let_name v).obj in
      fprintf' ppf "@[%s :@ %a@]@." (value_s name)
        (SyntacticTypesPrinting.print_polytype ~sugar_env:env.sugar) t;
      if !dump_svg then fprintf' ppf "@[%s@]@." (solved_constraint_to_svg c);
      TypingEnv.add_value env (name, t)

  | Inductive (({ obj = name}, params), constrs) ->
      let params' = TypingEnv.distinct_parameters params in
      TypingEnv.closed_inductive_declaration params' constrs.obj;
      TypingEnv.distinct_constructors constrs;
      let env' = TypingEnv.add_inductive env name params.obj constrs.obj in
      let name = TypeConstructors.typeconstr_s name in
      fprintf' ppf "@[Type %s@ defined.@ " name;
      List.iter (fun c -> fprintf' ppf "Constructor %s@ defined.@ "
                   (TypeConstructors.constr_s (fst c).obj)) constrs.obj;
      fprintf' ppf "Elimination@ scheme %s_elim@ defined.@]@." name;
      env'

  | Alias (({ obj = name }, params), t) ->
      let params' = TypingEnv.distinct_parameters params in
      TypingEnv.closed_polytype params' t;
      let env' = TypingEnv.add_alias env name params.obj t in
      fprintf' ppf "@[Type@ alias %s@ defined.@]@."
        (TypeConstructors.typeconstr_s name);
      env'

  | Inline (InlineArrow b) ->
      let sugar_arrow =
        if b then SyntacticTypes.sugar_arrow else [None; None] in
      { env with sugar =
          { env.sugar with SyntacticTypes.sugar_arrow = sugar_arrow }}

  | Inline (InlineUple s) ->
      { env with sugar =
          { env.sugar with SyntacticTypes.sugar_uple =
              SyntacticTypes.sugar_uple s }}

  | Inline (InlineConstr (c, ls)) ->
      add_sugar env c ls

  | Include f ->
      let f = String.uncapitalize f ^ ".mlf" in
      fprintf' ppf "@[Typing %s@]@." f;
      let env_options = get_env_options () in
      let r = try `Success (parse_and_type_file ?ppf:None ~env
                             ~catch_errors:false ~annot:false f)
      with e -> `Failure e
      in
      restore_env_options env_options;
      match r with
        | `Success e -> e
        | `Failure e -> raise e

and type_catch_phrase ?ppf env p =
  try (type_phrase ?ppf env p, true)
  with e ->
    ErrorsHandling.report_error (std_ppf ppf) e;
    (match e with
       | ConstraintsSolvingErrors.InferenceErrorReplayed (_, _, c) ->
           if !dump_svg then
             fprintf' ppf "@[%s@]@." (constraint_to_svg c)
       | _ -> ());
    (env, false)

and type_print_phrases ?ppf ?(catch_errors=true) env ?(prev_pos=0) input l =
  let rec aux env prev_pos = function
    | [] -> env
    | { obj = p; loc = l } :: q ->
        let pos_end = l.loc_end.Lexing.pos_cnum in
        reprint_source ?ppf prev_pos pos_end input;
        let env', outcome = type_catch_phrase ?ppf env p in
        let outcome =
          match p with Expr (_, false) -> not outcome | _ -> outcome in
        if outcome = false && catch_errors = false then failwith "TypePhrases";
        fprintf' ppf "@.";
        aux env' pos_end q
  in
  aux env prev_pos l

and parse_and_type_aux ?ppf ~env ~catch_errors input =
  try
    let p = Parser.phrases_ Lexer.token input.lexbuf in
    type_print_phrases ?ppf ~catch_errors env input p
  with e -> (* Intercept parsing errors *)
    ErrorsHandling.report_error (std_ppf ppf) e;
    env

and parse_and_type_string ?(ppf=std_formatter) ?(env=TypingEnv.initial_env) ?(highlight=true) ?(catch_errors=true) input =
    let input = {
      lexbuf = Lexing.from_string (Utils.add_end_newline input);
      source = input;
    } in
    if highlight then Highlight.input_lexbuf := Some input.lexbuf;
    parse_and_type_aux ~ppf ~env ~catch_errors input

and parse_and_type_file ?ppf ?(env=TypingEnv.initial_env) ?(catch_errors=true) ?(annot=false) f =
  Highlight.input_name := f;
  let ch = open_in f in
  let lexbuf = Lexing.from_channel ch in
  lexbuf.Lexing.lex_curr_p <-
    { lexbuf.Lexing.lex_curr_p with Lexing.pos_fname = f };
  let input = { lexbuf = lexbuf; source = "" } in
  if annot then (
    let ch_dump = open_out (Filename.chop_extension f ^ ".annot") in
    let ppf = Format.formatter_of_out_channel ch_dump in
    Dtypes.dump_ppf := Some ppf;
  );
  parse_and_type_aux ?ppf ~env ~catch_errors input
