open TypeConstructors
open SyntacticTypes
open KnownconstrTypes
open Locations
open TypingErrors



type 'a user_def =
  | Inductive of 'a constructor list
  | Alias of 'a

type 'a user_defined_type = {
  user_desc: user_type_desc;
  user_params: typevar list;
  user_def: 'a user_def
}


(* Typing environment. Types that cannot be overridden
   (arrow and tuples currently) are not mentionned *)
type env = {
  types: polytype user_defined_type list;
  constructors: (polytype * user_type_desc) MapConstrs.t;
  values: polytype AST.MapValues.t;
  sugar: sugar_env;
}

let add_value env (v, t) =
  { env with values = AST.MapValues.add v t env.values }

(* Find the most recent type constructor with the same name as
   the given type *)
let find_type_same_name ?(loc=dummy_loc) env t : overridable_typeconstr =
  try `User
    (List.find (fun t' -> t'.user_desc.typeconstr_name = t) env.types).user_desc
  with Not_found ->
    match typeconstr_s t with
      | "unit" -> `Unit
      | "int" -> `Int
      | "float" -> `Float
      | _ -> typing_error (UnknownTypeConstructor (t, loc))


(* Resolution of type constructors in the arguments *)

(* Resolves all constructors coming from a parsing phase using
   the given environment. Arities are also checked on the fly *)
let rec resolve_constrs loc env : UnknownconstrTypes.expanded_polytype -> KnownconstrTypes.expanded_polytype = function
  | `Var _ as v -> v
  | `Constr (c, l) ->
      let l' = List.map (resolve_constrs loc env) l
      and c' = match c with
        | `PConstr s ->
            let c = (find_type_same_name env ~loc s :> typeconstr) in
            if arity c <> List.length l then
              typing_error (IncorrectArity (s, arity c, List.length l, loc));
            c
        | #typeconstr as c -> c
      in
      `Constr (c', l')
  | `Bottom as b -> b
  | `Bound (l, (v, b), b') ->
      `Bound (l, (v, resolve_constrs loc env b), resolve_constrs loc env b')

let unexpand_resolve_constrs loc env (t : UnknownconstrTypes.expanded_polytype) =
  let t = resolve_constrs loc env t in
  unexpand env.sugar t



(* Type of a constructor, except for the return type which is the
   supplied monotype t *)
let constructor_type_aux (cargs : polytype list) (t : monotype) : polytype  =
  (* We convert the arguments of the constructor so that they are
     only expressed using monotypes *)
  let cargs' =
    let c = Utils.counter 1 in
    List.rev (List.map (function
                          | #monotype as m -> (m, None)
                          | p ->
                              let v = s_typevar (Printf.sprintf "_t%d" c#get) in
                              `Var v, Some (v, p))
                cargs)
  in
  (* Return type + the arguments, introduced using arrows *)
  let t = List.fold_left (fun t (m, _) -> arrow m t) t cargs' in
  (* Quantification for the arguments that were polytypes *)
  let t = List.fold_left
    (fun t (_, poly) -> match poly with
       | None -> t
       | Some (v, p) -> rigid v ~b:p t)
    (t :> _ polytype_ ) cargs' in
  t

(* A type constructor with its parameters *)
let type_with_params params t : monotype =
  `Constr (`User t, (List.map (fun v -> `Var v) params))

(* Bind the parameters of a type constructor in a polytype *)
let bind_params params t =
  List.fold_left (fun t v -> flex v t) t (List.rev params)


(* Find a fresh variable name (wrt. the names in params), made only of 'r' *)
let return_result_name params =
  let rec aux (name : typevar) =
    if List.mem name params then
      aux (s_typevar ((typevar_s name) ^ "r"))
    else
      name
  in aux (s_typevar "r")

(* Elimination principle of an inductive type *)
let elimination_principle params constrs t : polytype =
  (* Type to eliminate arrow result *)
  let v = return_result_name params in
  let r = `Var v in
  let tp = arrow (type_with_params params t) r in
  (* Type of the functions of each constructor *)
  let l = List.map (fun c -> constructor_type_aux c.constr_args r) constrs in
  (* Elimination principle except for the quantification *)
  let t = constructor_type_aux l tp in
  (* Parameters of the type *)
  let t = bind_params params t in
  (* Quantification for the return type *)
  flex v t

let elimination_principle_name name =
  AST.s_value (typeconstr_s name ^ "_elim")


(* Next available type descriptor for a type called 'name'. Returns
   a dummy description (as the definition of the type cannot be found
   immediately) and the environment enriched with the dummy type *)
let fresh_type_desc env name params poly =
  let fresh k = {
    typeconstr_name = name;
    typeconstr_name_id = k;
    typeconstr_arity = List.length params;
    typeconstr_is_poly = poly;
  }
  in
  try match find_type_same_name env name with
    | `User { typeconstr_name_id = k } -> fresh (k+1)
    | `Unit | `Int | `Float -> fresh 1
  with TypingError (UnknownTypeConstructor _) -> fresh 0


(* Add a new inductive to the typing environment. Any older type
   with the same name is shadowed (but not removed). The constructors
   are added to the known constructors, and the elimination principle
   is added to the list of knwon values *)
let add_inductive env name params (constrs : AST.inductive_constr_decl list) =
  let tdesc = fresh_type_desc env name params false in
  let dummy_type = {
    user_desc = tdesc;
    user_params = params;
    user_def = Inductive [] (* We cannot fill this list while constructors
                               are unresolved *)
  } in
  let env' = { env with types = dummy_type :: env.types } in
  (* Resolution of type constructors in a constructor. We also
     "unexpand" the expanded_polytype into a real polytype *)
  let resolve_constrs c =
    unexpand_resolve_constrs c.loc env' c.obj
  in
  (* Resolution of a type constructor in the entire inductive *)
  let constrs' = List.map
    (fun (name, args) -> {
       constr_name = name.obj;
       constr_args = List.map resolve_constrs args;
     }) constrs
  in
  (* We update the environment with the real type definition *)
  let real_type = { dummy_type with user_def = Inductive constrs' } in
  let env' = { env' with types = real_type :: List.tl env'.types } in

  let constructors_types =
    let return_type = type_with_params params tdesc in
    List.map
      (fun c -> c.constr_name,
         bind_params params (constructor_type_aux c.constr_args return_type))
      constrs'
  and elim = elimination_principle params constrs' tdesc
  in
  { env' with
      constructors =
        List.fold_left (fun s (c, ct) -> MapConstrs.add c (ct, tdesc) s)
          env'.constructors constructors_types;
      values = AST.MapValues.add (elimination_principle_name name)
        elim env'.values;
  }

(* Similarly, add a type alias to the environment *)
let add_alias env name params t =
  let t' = unexpand_resolve_constrs t.loc env t.obj in
  let poly = polymorphic_type t' in
  let tdesc = fresh_type_desc env name params poly in
  let t_alias = {
    user_desc = tdesc;
    user_params = params;
    user_def = Alias t';
  } in
  let env' = { env with types = t_alias :: env.types } in
  env'


(* Check that all type parameters of a type are distinct,
   and return them as a set *)
let distinct_parameters l =
  List.fold_left (fun s var ->
                    if SetTypeVars.mem var s then
                      typing_error (NonDistinctParameters (var, l.loc))
                    else
                      SetTypeVars.add var s)
    SetTypeVars.empty l.obj


(* Check that a polytype is closed wrt to a set of type variables *)
let closed_polytype params t =
  let ftv = ftv' t.obj in
  let diff = SetTypeVars.diff ftv params in
  if SetTypeVars.is_empty diff = false then
    typing_error (UnboundTypeVar (SetTypeVars.min_elt diff, t.loc))

(* Check that all arguments of a constructor are closed (ie. their
   ftv are a subset of the parameters of the type) *)
let closed_constructor_declaration params (c : AST.inductive_constr_decl) =
  List.iter (closed_polytype params) (snd c)

(* Check that all constructors declarations in a type declaration are closed *)
let closed_inductive_declaration params constrs =
  List.iter (closed_constructor_declaration params) constrs


(* Check that all constructors declared are distinct *)
let distinct_constructors (l : AST.inductive_constr_decl list loc) =
  ignore (List.fold_left (fun constrs ({ obj = c }, _) ->
                            if SetConstrs.mem c constrs then
                              typing_error (NonDistinctConstructors (c, l.loc))
                            else
                              SetConstrs.add c constrs)
            SetConstrs.empty l.obj)

let add_sugar env c ls =
  match find_type_same_name env c with
    | #init_ct_typeconstr ->
        typing_error (IncorrectSugarAnnotation (c, 0, ls.loc))
    | `User d ->
        if List.length ls.obj <> d.typeconstr_arity then
          typing_error (IncorrectSugarAnnotation (c, d.typeconstr_arity, ls.loc))
        else
          { env with sugar = { env.sugar with sugar_inductive =
                (d, ls.obj) :: env.sugar.sugar_inductive }}


(* Helper functions to add types to the initial typing envrironment *)
let builtin_inductive (name, params, constrs) =
  s_typeconstr name,
  params,
  List.map (fun (name, args) ->
              (noloc (s_constr name), List.map noloc args)) constrs

let builtin_alias (name, params, t) =
  s_typeconstr name,
  params,
  noloc t

let bool = "bool", [], ["True", [];
                        "False", []]

let list =
  let v = s_typevar "a" in
  "list", [v], ["Nil", [];
                "Cons", [`Var v;
                         `Constr (`PConstr (s_typeconstr  "list"), [`Var v])]]

let t =
  let v = s_typevar "a" in
  "t", [v], `Constr (`Int, [])

let church_int =
  let v = s_typevar "a" in
  "cint", [],
  flex v (arrow (arrow (var "a") (var "a")) (arrow (var "a") (var "a")))

let f3uple f f' c env = let (c1, c2, c3) = f' c in f env c1 c2 c3
let add_inductive' = f3uple add_inductive builtin_inductive
let add_alias' = f3uple add_alias builtin_alias


let initial_env =
  let (>=) x f = f x in
  let init = {
    types = [];
    constructors = MapConstrs.empty;
    values = AST.MapValues.empty;
    sugar = default_sugar_env;
  } in
    init
    >= add_inductive' bool
(*    >= add_inductive' list *)
(*  >= add_alias' t;
    >= add_alias' church_int; *)


let initial_ct s =
  unexpand_resolve_constrs dummy_loc initial_env
    (Parser.polytype_ Lexer.token (Lexing.from_string s))



let initial_env =
  let (>=) env (v, t) = add_value env (AST.s_value v, initial_ct t) in
  initial_env
  >= ("incr", "int -> int")
  >= ("decr", "int -> int")
  >=  ("assert", "'a. bool -> 'a")
  >= ("either", "'a. 'a -> 'a -> 'a")
  >= ("ignore", "'a. 'a -> unit")
  >= ("band", "bool -> bool -> bool")
  >= ("fix", "'a. ('a -> 'a) -> 'a")
  >= ("fst", "'a. 'b. 'a * 'b -> 'a")
  >= ("snd", "'a. 'b. 'a * 'b -> 'b")
