(*pp camlp4o ./hashtbl_ext.cmo *)

open SyntacticTypes
open GraphicTypes
open GraphicTypes.Nodes
open Utils
open GraphicTypesAux
open TypeConstructors
open TypingEnv
open Format

let () = #{HashNodes}

let (!!!) n = !!n.unif_info

let current_run u n =
  Tag.equal !!!n.unif_tag u

let reset_unif_info tag ni =
  let ui = ni.unif_info in
  if Tag.equal ui.unif_tag tag = false then (
    ui.unif_tag <- tag;
    ui.unified_nodes <- Tree.Leaf (ni.id, ni.bound);
    ui.bot_non_bot <- false;
    ui.merge_took_place <- false;
  )

exception ContainsConstraints

type unif_error =
  | Cyclic
  | Clash of (node * typeconstr) * (node * typeconstr)
  | ForbiddenGraft of NId.id * NId.id list
  | ForbiddenRaise of NId.id * NId.id list
  | ForbiddenWeaken of NId.id * NId.id list
  | ForbiddenMerge of NId.id * NId.id list

let print_unif_error ?(output=(!Options.output)) ppf = function
  | Cyclic -> fprintf ppf "cyclic@ type"
  | ForbiddenGraft _ -> fprintf ppf "forbidden@ graft"
  | ForbiddenRaise _ -> fprintf ppf "forbidden@ raise"
  | ForbiddenWeaken _ -> fprintf ppf "forbidden@ weaken"
  | ForbiddenMerge _ -> fprintf ppf "forbidden@ merge"
  | Clash ((_, c1), (_, c2)) ->
      fprintf ppf "constructor@ clash@ between@ %s@ and@ %s"
        (string_of_typeconstr ~output c1)
        (string_of_typeconstr ~output c2)

let print_unif_error_explanation ?(output=(!Options.output)) s ppf = function
  | Cyclic -> fprintf ppf "the@ resulting@ type@ would@ be@ cyclic"
  | Clash ((_, c1), (_, c2)) -> fprintf ppf
      "constructor@ clash@ between@ %s@ and@ %s"
        (string_of_typeconstr ~output c1)
        (string_of_typeconstr ~output c2)
  | ForbiddenGraft _
  | ForbiddenRaise _
  | ForbiddenWeaken _
  | ForbiddenMerge _ as e ->
      fprintf ppf "the@ %s@ is@ probably@ not@ polymorphic@ enough@ (%a)"
        s (print_unif_error ~output) e

(* Standard unification errors *)
exception UnifError of unif_error

let unif_error e = raise (UnifError e)


let extract_ids n =
  Tree.fold (fun (id, _) l -> id :: l) [] !!!n.unified_nodes


(* Expansion of abbreviations *)

let () = #{Hashtbl} (* Used on typevar *)

let expand_alias_aux n (params : typevar list) t =
  let env = {} in
  List.iter2 (Hashtbl.add env) params (children n);
  let n' = GraphicSyntacticTranslations.polytype_to_graph
    ~env ~perms:(permissions n) ~bheight:(binding_height n + 1) t in
  (* If the type is a completely degenerate annotation that does not
     create a new node, we need to keep the binder of the variable.
     We do this strange test to avoid stupid abbreviations
     such as type 'a t = 'b. 'a.
  *)
  if List.exists (fun n'' -> Nodes.equal n' n'' &&
                    Nodes.equal n (binder n'') = false) (children n) then
    fuse ~keep:n' ~lose:n ~bound:n' ()
  else
    fuse ~keep:n' ~lose:n ~bound:n ()

let expand_alias env n =
  match !!n.constr with
    | Bot -> false
    | Forall _ -> raise ContainsConstraints
    | Type t -> expand_alias_aux n [] t; true
    | Constr `User d ->
        let t = List.find (fun t -> equal_user_typedesc d t.user_desc)
          env.types in
        (match t.user_def with
           | Inductive _ -> false
           | Alias t' -> expand_alias_aux n t.user_params t'; true
        )
    | Constr _ -> false

let expand_type n = match !!n.constr with
  | Bot | Forall _ | Constr _ -> ()
  | Type t -> expand_alias_aux n [] t

let expand_after_clash env n1 n2 =
  expand_alias env n1 || expand_alias env n2


let () = #{HashNodes}

let unif_struct tenv tag (n1 : node) (n2 : node) =
  let list_bot_non_bot = new store ()
  and check_graft = new store () in

  (* Fusion of two nodes. We keep the informations on the
     nodes merged, and we store the fact that 2 nodes
     were merged *)
  let fuse ~keep ~lose = UnionFind.union
    ~update:(fun ~keep:nikeep ~lose:nilose ->
       nikeep.unif_info.unified_nodes <-
         Tree.Node (nikeep.unif_info.unified_nodes,
                    nilose.unif_info.unified_nodes);
       nikeep.unif_info.merge_took_place <- true;
    ) ~keep:keep.node_info ~lose:lose.node_info ()

  (* Memorises the fact that a bot and a non bottom node
     has been unified (which will be used to add virtual binding
     edges *)
  and add_bot_non_bot n ui =
    if ui.bot_non_bot = false then (
      ui.bot_non_bot <- true;
      list_bot_non_bot#add n)
  in

  let rec aux (n1 :node) (n2 : node) =
    if not (UnionFind.equivalent n1.node_info n2.node_info) then
      let i1 = !!n1 and i2 = !!n2 in
      reset_unif_info tag i1;
      reset_unif_info tag i2;
      match i1.constr, i2.constr with
        | Bot, Bot ->
            fuse ~keep:n1 ~lose:n2
        | _, Bot ->
            check_graft#add (n1, i2.unif_info.unified_nodes);
            add_bot_non_bot n1 i1.unif_info;
            fuse ~keep:n1 ~lose:n2
        | Bot, _ ->
            check_graft#add (n1, i1.unif_info.unified_nodes);
            add_bot_non_bot n2 i2.unif_info;
            fuse ~keep:n2 ~lose:n1
        | Constr c1, Constr c2 ->
            if c1 <> c2 then (
              if expand_after_clash tenv n1 n2 then
                aux n1 n2
              else
                unif_error (Clash ((n1, c1), (n2, c2))))
            else (
              fuse ~keep:n1 ~lose:n2;
              Utils.iter2 aux i1.children i2.children;
            )
        | Forall _, _ | _, Forall _ -> raise ContainsConstraints
        | Type _, Type _ -> expand_type n1; expand_type n2; aux n1 n2
        | Type _, _ -> expand_type n1; aux n1 n2
        | _, Type _ -> expand_type n2; aux n1 n2
  in
  aux n1 n2;
  if dfs_struct.has_cycle n1 then unif_error Cyclic;
  (* We do the checks for graft after the cyclicity test
     to avoid some weird error messages *)
  List.iter (fun (n, t) -> Tree.iter
               (fun (id, bound) -> match bound with
                  | Root | Bound { permissions = PFlex } -> ()
                  | Bound _ -> unif_error (ForbiddenGraft (id, extract_ids n)))
               t)
    check_graft#get;
  (* Nodes resulting from the merge of a bot and a non bottom node
     are used to add virtual edges *)
  list_bot_non_bot#get


(* Invariant : virtual_edges{n} exists as soon as n has been visited *)
let add_virtual_edges virtual_edges tag bot_non_bot =

  let rec update_node ?(b=None) (n : node) =
    let ni = !!n in
    reset_unif_info tag ni;
    let already_visited, virtual_edges_n =
      try true, virtual_edges{n}
      with Not_found -> false, []
    in

    (* If b is not None, we add the binding edge to n *)
    (match b with
      | None -> ()
      | Some b -> virtual_edges{n} <- b :: virtual_edges_n);

    (* We mark the fact we have visited n if needed *)
    if b = None && already_visited = false then
      virtual_edges{n} <- [];

    (* If n has not already been visited, we compute the virtual
       edges for its children *)
    if already_visited = false then
      List.iter (update_node ~b:(Some n)) (children n)
  in
  List.iter update_node bot_non_bot


let rebind tag virtual_edges n =
  let nodes = dfs_struct.topological_sort_list n in
  let update_node n =
    let b1 = Tree.fold (fun (id, bound) l ->
                          match bound with
                            | Root -> l
                            | Bound b -> (id, b) :: l) [] !!!n.unified_nodes
    and b2 = try virtual_edges{n} with Not_found -> []
    in
    if b1 <> [] or b2 <> [] then
      (* Flag *)
      let new_flag =
        let rec aux = function
          | [] -> Flex
          | (_, { label = Rigid }) :: _ -> Rigid
          | (_, { label = Flex }) :: q -> aux q
        in aux b1
      in

      (* Check for the permissions of the weakenings *)
      if new_flag = Rigid then
        List.iter (fun (id, {label = l; permissions = p}) ->
                     if l = Flex && p <> PFlex then
                       unif_error (ForbiddenWeaken (id, extract_ids n))) b1;

      (* Computation of the binder *)
      let new_binder =
        (* Ancestors contain the binders of all the nodes of t_g
           merged in n *)
        let ancestors = SetNodesHeight.union
          (List.fold_left (fun l (_, {binder = n}) -> SetNodesHeight.add n l)
             SetNodesHeight.empty b1)
          (List.fold_left (fun l n -> SetNodesHeight.add n l)
             SetNodesHeight.empty b2)
        in
        lca_binders ancestors
      in
      (* Check the permissions for the raisings *)
      List.iter (fun (id, {binder = n'; permissions = p}) ->
                   if Nodes.equal n' new_binder = false && p = PLocked
                   then unif_error (ForbiddenRaise (id, extract_ids n))) b1;


      (* Stores the new bound *)
      let binder_perms, binder_binding_height = match !!new_binder.bound with
        | Root -> PFlex, 0
        | Bound b -> b.permissions, b.binding_height
      in
      set_binder n ({
        label = new_flag;
        binder = new_binder;
        permissions = permissions_syst binder_perms new_flag;
        binding_height = binder_binding_height + 1
      });
  in
  List.iter (fun n -> if Tag.equal !!!n.unif_tag tag then update_node n) nodes

module HashId = Hashtbl.Make(NId)
let _ = #{HashId}

let build_old_binder n1 n2 =
  let update = new store () in
  (* Maps the node after unification to their old self. A mapping
     for a node will be silently updated during the walk *)
  let new_node_to_old_node = {}
  (* Maps an old node to its old binder in \Gcs. A mapping
     is never updated *)
  and old_binder = {} in
  let walk =
    (* Used for the dfs *)
    let h = {} in
    let rec walk n =
      if !!!n.merge_took_place && h{? !!.n.old_id} = false then (
        h{!!.n.old_id} <- ();
        update#add n;

        (* We update the mapping for the current new node, in
           case it is used to bind a node *)
        new_node_to_old_node{!!n.id} <- !!.n.old_id;

        (* We compute the old binder *)
        (match !!n.bound with
          | Root -> ()
          | Bound {binder = b} ->
              old_binder{!!.n.old_id} <-
                (try new_node_to_old_node{!!b.id}
                 with Not_found ->
                   (* Not found cases can happen if the node is bound
                      above the root of unification. In this case, no
                      merge has happened, and old_id and the new id are
                      still equal *)
                   assert (NId.equal !!.b.old_id !!b.id);
                   !!.b.old_id);
        );

        List.iter walk !!.n.old_children;

        new_node_to_old_node{!!n.id} ->;

      )
    in walk
  in
  walk n1;  walk n2;
  (* We update the old structure, it won't be used anymore *)
  List.iter (fun n -> UnionFind.union ~keep:!!n.canonical_node.old_struct
               ~lose:n.old_struct ())
    update#get;
  old_binder


(* Test merges for the new node n *)
let test_merges_node old_binder n =
  let h = {} in
  Tree.iter
    (fun (id, b) ->
       match b with
         | Root -> assert false
         | Bound {permissions = p} ->
             let n' = old_binder{id} in
             h{n'} <- (id, p) :: (try h{n'} with Not_found -> [])
    ) !!!n.unified_nodes;
  let test_merge l =
    if List.length l >= 2 then
      List.iter (fun (id, p) ->
                   if p = PLocked then
                     unif_error (ForbiddenMerge (id, extract_ids n))) l
  in HashId.iter (fun _id l -> test_merge l) h

let () = #{HashNodes}

(* Test merges for all the nodes in which a merge took place *)
let test_merges tag old_binder n =
  let visited = {} in
  let rec aux n =
    if visited{?n} = false &&
      Tag.equal !!!n.unif_tag tag && !!!n.merge_took_place then (
        visited{n} <- ();
        test_merges_node old_binder n;
        List.iter aux (children n);
      )
  in aux n


let unif env (n1, n2) =
  let tag = Tag.tag () in
  let bot_non_bot = unif_struct env tag n1 n2 in
  let virtual_edges = {} in
  add_virtual_edges virtual_edges tag bot_non_bot;
  rebind tag virtual_edges n1;
  let old_binder = build_old_binder n1 n2 in
  test_merges tag old_binder n1;
