(*pp camlp4o ./hashtbl_ext.cmo *)

open GraphicTypes
open GraphicTypes.Nodes
open GraphicTypesAux
open Constraints
open Utils

let () = #{HashNodes}

type nid = int

exception HasCycle of node
exception BoundRoot
exception RootNonRoot of node
exception IncorrectArity of node
exception Domination of nid
exception Permissions of node
exception BindingHeight of nid * int
exception OldStructure of node
exception ForallWf of nid
exception ForallArity of node * int * int

let correct c =
  let r = c.root in
  let succs n = constraints_succs c n in
  let dfs = aux_dfs succs in
  let tarjan_lengauer = tarjan succs r in
  let rec dominates n n' =
    Nodes.equal n n' ||
      match tarjan_lengauer n' with
        | None -> false
        | Some n' -> dominates n n'
  in

  let rec acyclic r =
    if dfs.has_cycle r then raise (HasCycle r)

  and is_root n =
    if !!n.bound <> Root then raise BoundRoot

  and non_root_node n =
    if Nodes.equal r n = false && !!n.bound = Root then raise (RootNonRoot n)

  and arity n =
    if List.length !!n.children <> garity !!n.constr then
      raise (IncorrectArity n)

  and binder_dominates n =
    match !!n.bound with
      | Root -> ()
      | Bound { binder = n' } ->
          if dominates n' n = false then
            raise (Domination (nid n))

  and correct_permission n =
    let rec node_permission n =
      match !!n.bound with
        | Root -> PFlex
        | Bound {binder = n'; label = l} ->
            permissions_syst (node_permission n') l
    in
    match !!n.bound with
      | Root -> ()
      | Bound {permissions = p} ->
          if p <> node_permission n then raise (Permissions n)

  and correct_binding_height n =
    match !!n.bound with
      | Root -> ()
      | Bound {binding_height = h; binder = n'} ->
          if h <> binding_height n' + 1
          then raise (BindingHeight (nid n, h))

  and correct_old n =
    if NId.equal !!.n.old_id !!n.id = false ||
      List.length !!.n.old_children <> List.length !!n.children ||
      List.for_all2 Nodes.equal !!.n.old_children !!n.children = false
    then raise (OldStructure n)

  and correct_forall { edge_type = ct ; edge_nodes = (f, n) } =
    if ct = Constraints.Inst then
      let forall_ok =
        match !!f.bound with
          | Root -> true
          | Bound { permissions = p } -> p = PFlex
      and n_ok = match !!n.bound with
        | Root -> true
        | Bound { binder = n'; } -> permissions n' = PFlex
      in
      if forall_ok = false || n_ok = false
        (*|| (Nodes.equal (binder (child1 f)) f &&
          (transitively_bound (binder f) n = false)) *)
      then raise (ForallWf (nid f))

  and correct_arity_forall n =
    match !!n.constr with
      | Forall { forall = i } ->
          let i' =
            List.fold_left (fun i' { edge_nodes = (n', _) } ->
                              if Nodes.equal n n' then i'+1 else i')
              0  c.constraints#get
          in
          if i' <> i then raise (ForallArity (n, i, i'))
      | _ -> ()

  in
  acyclic r;
  is_root r;
  dfs.generic_dfs
    ~pre:(fun n ->
            non_root_node n;
            arity n;
            correct_old n;
            correct_permission n;
            correct_binding_height n;
            correct_arity_forall n;
            binder_dominates n;
         ) r;
  List.iter correct_forall c.constraints#get;
  true
