(* Code by Francois Pottier, in which I inlined the signature of
   the argument of the functor *)


(* This module provides an implementation of Lengauer and Tarjan's
   algorithm for finding dominators in a flow graph.

   The algorithm runs when the functor is applied. Its complexity is
   essentially $O(E)$, where $E$ is the number of edges in the graph
   $G$. *)

module type G = sig

  type node

  (* We assume each node has a unique index. Indices must range from
     $0$ to $n-1$, where $n$ is the number of nodes in the graph. *)

  val n: int
  val index: node -> int

  (* Given a node, this function allows iterating over its immediate
     successors. *)

  val successors: (node -> unit) -> node -> unit

  (* Given a node, this function allows iterating over its immediate
     predecessors. *)

  val predecessors: (node -> unit) -> node -> unit

  (* A start node is a node from which every node is reachable. *)

  val start: node

end


module Run (G : G) : sig

  open G

  (* Given a node other than the start node, this function returns its
     immediate dominator. Given the start node, it returns [None]. *)

  val dominator: node -> node option

  (* Given a node, this function returns the number assigned to it by
     the initial depth-first traversal of the graph.  (Not to be
     confused with a depth-first traversal of the dominator tree!)

     This information can be useful when comparing nodes which are
     known to belong to the same branch of the dominator tree. Indeed,
     if [x] and [y] are known to be comparable in the dominance
     relation, then [x] (strictly) dominates [y] if and only if [x]'s
     number is (strictly) smaller than [y]'s. *)

  val number: node -> int

end

